'''
2020-2021, David Rowenhorst/The US Naval Research Laboratory, Washington DC
# Pursuant to title 17 section 105 of the United States Code, works of US Governement employees 
# are not not subject to copyright protection.
#
# Copyright (c) 2013-2014, Marc De Graef/Carnegie Mellon University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are
# permitted provided that the following conditions are met:
#
#     - Redistributions of source code must retain the above copyright notice, this list
#        of conditions and the following disclaimer.
#     - Redistributions in binary form must reproduce the above copyright notice, this
#        list of conditions and the following disclaimer in the documentation and/or
#        other materials provided with the distribution.
#     - Neither the names of Marc De Graef, Carnegie Mellon University nor the names
#        of its contributors may be used to endorse or promote products derived from
#        this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
####################################################################################################


#Author David Rowenhorst/The US Naval Research Laboratory, Washington DC
# A translation of the original f.90 code produced by Marc De Graef
#
# This implementation is dependant on the Numba Python package, which gives a significant increase in 
# speed (other than the first time called due to compilation time). The numba implmentation is very non-Numpy-like in that
# loops are not discouraged.
# No guarantee - but this library is typically numerically accurate to around 5e-6 degrees or better.
# The rotations.py is typically more accurate at the cost of being about 2-10x slower.

# All inputs can be numpy array-like, and either arrays of shape (n,m) OR (m)  where m is the length of a single 
# representation (i.e. for Euler m = 3, for quaternions m = 4); in the case of orientation matrices (om)
# the inputs/outputs should be (n,3,3) OR (3,3). 
# 
# If the inputs are not in C-order or not aligned, a copy of the input will be made. 
#
#
'''

import numpy as np
import numba
from os import environ
import tempfile
from pathlib import PurePath
import platform
tempdir = PurePath("/tmp" if platform.system() == "Darwin" else tempfile.gettempdir())
tempdir = tempdir.joinpath('numba')
environ["NUMBA_CACHE_DIR"] = str(tempdir)

P = 1
eps = 1.0e-12 # used for "closeto" approximations in Numba code
PI = np.pi
PI2 = 2.0 * np.pi
nbcache = True # switch to false for debugging.
nbParallel = False # decide if parallelism is desired.
# Once the JIT compiler runs over the code, changing this will not change the parallelism.  
# The package will need to be reimported
nbFastmath = True


# this will check the type of the input and its size.  
# it will also reshape the data in case it is a single dimension
# returns a 2D view of the data, the length of one entry, the number of entries, and datatype. 
@numba.jit(nopython=True,fastmath=nbFastmath,cache=nbcache)
def prepIn(a):
  shape = a.shape
  intype = a.dtype
  m = shape[-1]
  n = numba.int64(a.size / m)
  aout = a.reshape((n,m))
  return (aout,m,n,intype)


def eu2om(eu,p=P):
  '''Bunge Euler angles to orientation matrix'''
  om = eu2omL(np.require(eu,requirements=['C','A']),p=p)
  om = np.squeeze(om)
  return om


@numba.jit(nopython=True,fastmath=nbFastmath,cache=nbcache,parallel=nbParallel)
def eu2omL(euIn,p=P):
  eu,m,n,intype = prepIn(euIn)
  om = np.zeros((n,3,3),dtype=intype)
  for i in numba.prange(n):
    c1 = np.cos(eu[i,0])
    c2 = np.cos(eu[i,1])
    c3 = np.cos(eu[i,2])
    s1 = np.sin(eu[i,0])
    s2 = np.sin(eu[i,1])
    s3 = np.sin(eu[i,2])
    om[i,0,0] = c1 * c3 - s1 * s3 * c2
    om[i,0,1] = s1 * c3 + c1 * s3 * c2
    om[i,0,2] = s3 * s2
    om[i,1,0] = -c1 * s3 - s1 * c3 * c2
    om[i,1,1] = -s1 * s3 + c1 * c3 * c2
    om[i,1,2] = c3 * s2
    om[i,2,0] = s1 * s2
    om[i,2,1] = -c1 * s2
    om[i,2,2] = c2
  return om

 
def eu2ax(eu,p=P):
  '''Bunge Euler angles to axis-angle (radians)  '''
  ax = eu2axL(np.require(eu,requirements=['C','A']),p=p)
  ax = np.squeeze(ax)
  return ax


@numba.jit(nopython=True,fastmath=nbFastmath,cache=nbcache,parallel=nbParallel)
def eu2axL(euIn,p=P):
  pf = numba.float32(p > 0) * 2.0 - 1.0
  eu,m,n,intype = prepIn(euIn)
  ax = np.zeros((n,4),dtype=intype)
  for i in numba.prange(n):
    t = np.tan(eu[i,1] * 0.5)
    sig = 0.50 * (eu[i,0] + eu[i,2])
    ssig = np.sin(sig)
    delt = 0.50 * (eu[i,0] - eu[i,2])
    tau = np.sqrt(t * t + ssig * ssig)

    if (np.abs(sig - PI2) < eps):
      alpha = PI
    else:
      alpha = 2.0 * np.arctan(tau / np.cos(sig))

    if np.abs(alpha) < eps:
      ax[i,0] = 0.0
      ax[i,1] = 0.0
      ax[i,2] = -1.0 * pf
      ax[i,3] = 0.0
    else:
      ax[i,0] = -1.0 * pf * t * np.cos(delt) / tau
      ax[i,1] = -1.0 * pf * t * np.sin(delt) / tau
      ax[i,2] = -1.0 * ssig / tau
      ax[i,3] = alpha

    if alpha < 0.0:
      for j in range(4):
        ax[i,j] *= -1.0
  return ax


def eu2qu(eu,p=P):
  '''Bunge Euler angles to quaternions (scalar, axis)  '''
  qu = eu2quL(np.require(eu,requirements=['C','A']),p=p)
  qu = np.squeeze(qu)
  return qu


@numba.jit(nopython=True,fastmath=nbFastmath,cache=nbcache,parallel=nbParallel)
def eu2quL(euIn,p=P):
  pf = numba.float32(p > 0) * 2.0 - 1.0
  eu,m,n,intype = prepIn(euIn)
  qu = np.zeros((n,4),dtype=intype)
  for i in numba.prange(n):
    phi1 = eu[i,0] * 0.5
    phi = eu[i,1] * 0.5
    phi2 = eu[i,2] * 0.5

    dphi1phi2 = phi1 - phi2
    sphi1phi2 = phi1 + phi2
    cphi = np.cos(phi)
    sphi = -1.0 * np.sin(phi)
    qu[i,0] = cphi * np.cos(sphi1phi2) + 0.0
    sign = 1.0 if (qu[i,0] >= 0.0) else -1.0
    qu[i,0] *= sign
    qu[i,1] = pf * sign * sphi * np.cos(dphi1phi2) + 0.0
    qu[i,2] = pf * sign * sphi * np.sin(dphi1phi2) + 0.0
    qu[i,3] = -1.0 * pf * sign * cphi * np.sin(sphi1phi2) + 0.0
  qu = quatnormL(qu)
  return qu


def om2eu(om,p=P):
  '''orientation matrix to Bunge Euler angles'''
  if om.ndim == 3:
    omIn = np.require(om,requirements=['C','A'])
  elif om.ndim == 2:
    omIn = np.require(om,requirements=['C','A']).reshape((1,3,3))
  else:
    return -1
  eu = om2euL(omIn,p=p)
  eu = np.squeeze(eu)
  return eu


@numba.jit(nopython=True,fastmath=nbFastmath, cache=nbcache, parallel=nbParallel)
def om2euL(om, p=P):
  pf = numba.float32(p > 0) * 2.0 - 1.0
  intype = om.dtype
  n = np.int64(om.size / 9)
  eu = np.zeros((n, 3), dtype=intype)
  for i in numba.prange(n):
    # [0,0], [0,1], [0, 2], [1,0], [1,1], [1,2], [2,0], [2,1], [2,2]
    #    0      1      2      3      4      5      6      7      8
    if (1.0 - np.abs(om[i, 2, 2])) > eps:
      eu[i,1] = np.arccos(om[i,2,2])
      zeta = 1.0 / (np.sqrt(1.0 - om[i,2,2] * om[i,2,2]))
      eu[i,0] = np.arctan2(om[i,2,0] * zeta,-1.0 * om[i,2,1] * zeta)
      eu[i,2] = np.arctan2(om[i,0,2] * zeta,om[i,1,2] * zeta)
    else:
      if ((np.abs(om[i,2,2] - 1.0)) < eps):
        eu[i,0] = np.arctan2(om[i,0,1],om[i,0,0])
        eu[i,1] = 0.0
        eu[i,2] = 0.0
      else:
        eu[i,0] = -np.arctan2(om[i,0,1],om[i,0,0])
        eu[i,1] = PI
        eu[i,2] = 0.0
    eu[i,0] += 100.0 * PI
    eu[i,0] = np.fmod(eu[i,0],PI2)
    eu[i,2] += 100.0 * PI
    eu[i,2] = np.fmod(eu[i,2],PI2)
  return eu


def ax2om(ax,p=P):
  '''axis-angle (radians) to orientation matrix  '''
  om = ax2omL(np.require(ax,requirements=['C','A']),p=p)
  om = np.squeeze(om)
  return om


@numba.jit(nopython=True,fastmath=nbFastmath,cache=nbcache,parallel=nbParallel)
def ax2omL(axIn,p=P):
  ax,m,n,intype = prepIn(axIn)
  om = np.zeros((n,3,3),dtype=intype)
  axn = axnormL(ax)
  # [0,0], [0,1], [0, 2], [1,0], [1,1], [1,2], [2,0], [2,1], [2,2]
  #    0      1      2      3      4      5      6      7      8
  for i in numba.prange(n):
    c = np.cos(axn[i,3])
    s = np.sin(axn[i,3])
    omc = 1.0 - c
    for j in range(3):
      om[i,j,j] = axn[i,j] * axn[i,j] * omc + c
    q = omc * axn[i,0] * axn[i,1]
    om[i,1,0] = q + s * axn[i,2]
    om[i,0,1] = q - s * axn[i,2]

    q = omc * axn[i,1] * axn[i,2]
    om[i,2,1] = q + s * axn[i,0]
    om[i,1,2] = q - s * axn[i,0]

    q = omc * axn[i,2] * axn[i,0]
    om[i,0,2] = q + s * axn[i,1]
    om[i,2,0] = q - s * axn[i,1]

    if (p < 0):
      om[i,:,:] = np.transpose(om[i,:,:])

  return om


def qu2eu(qu,p=P):
  '''quaternions (scalar, axis) to Bunge Euler angles  '''
  eu = qu2euL(np.require(qu,requirements=['C','A']),p=p)
  eu = np.squeeze(eu)
  return eu


@numba.jit(nopython=True,fastmath=nbFastmath,cache=nbcache,parallel=nbParallel)
def qu2euL(quIn,p=P):
  pf = numba.float32(p > 0) * 2.0 - 1.0
  qu,m,n,intype = prepIn(quIn)
  eu = np.zeros((n,3),dtype=intype)
  for i in numba.prange(n):
    q0 = pf * qu[i,0]
    q03 = q0 * q0 + qu[i,3] * qu[i,3]
    q12 = qu[i,1] * qu[i,1] + qu[i,2] * qu[i,2]
    chi = np.sqrt(q03 * q12)

    if (chi == 0.0):
      if (q12 == 0.0):
        eu[i,2] = 0.0
        eu[i,1] = 0.0
        eu[i,0] = np.arctan2(-2.0 * q0 * qu[i,3],q0 * q0 - qu[i,3] * qu[i,3])
      else:
        eu[i,2] = 0.0
        eu[i,1] = PI
        eu[i,0] = np.arctan2(2.0 * qu[i,1] * qu[i,2],qu[i,1] * qu[i,1] - qu[i,2] * qu[i,2])
    else:
      eu[i,1] = np.arctan2(2.0 * chi,q03 - q12)

      diff = np.arctan2(-qu[i,2],-qu[i,1])
      sum = np.arctan2(-qu[i,3],q0)
      eu[i,0] = diff + sum
      eu[i,2] = sum - diff

    eu[i,0] += PI2
    eu[i,0] = np.fmod(eu[i,0],PI2)
    eu[i,2] += PI2
    eu[i,2] = np.fmod(eu[i,2],PI2)

  return eu


def ax2ho(ax,p=P):
  '''axis-angle (radians) to homochoric vector  '''
  ho = ax2hoL(np.require(ax,requirements=['C','A']))
  ho = np.squeeze(ho)
  return ho


@numba.jit(nopython=True,fastmath=nbFastmath,cache=nbcache,parallel=nbParallel)
def ax2hoL(axIn,p=P):
  pf = numba.float32(p > 0) * 2.0 - 1.0
  # intype = ax.dtype
  # n = np.int64(ax.size / 4)
  ax,m,n,intype = prepIn(axIn)
  ho = np.zeros((n,3),dtype=intype)
  axn = axnormL(ax)
  #axn = ax
  for i in numba.prange(n):
    f = 0.75 * (axn[i,3] - np.sin(axn[i,3]))
    f = f ** (1.0 / 3.0)
    for j in range(3):
      ho[i,j] = f * axn[i,j]
  return ho


def ho2ax(ho,p=P):
  '''homochoric vector to axis-angle (radians)   '''
  ax = ho2axL(np.require(ho,requirements=['C','A']))
  ax = np.squeeze(ax)
  return ax


@numba.jit(nopython=True,fastmath=nbFastmath,cache=nbcache,parallel=nbParallel)
def ho2axL(hoIn,p=P):
  pf = numba.float32(p > 0) * 2.0 - 1.0
  ho,m,n,intype = prepIn(hoIn)
  ax = np.zeros((n,4),dtype=intype)

  tfit = np.array([1.0000000000018852,-0.5000000002194847,-0.024999992127593126,
                   -0.003928701544781374,-0.0008152701535450438,-0.0002009500426119712,
                   -0.00002397986776071756,-0.00008202868926605841,0.00012448715042090092,
                   -0.0001749114214822577,0.0001703481934140054,-0.00012062065004116828,
                   0.000059719705868660826,-0.00001980756723965647,0.000003953714684212874,
                   -0.00000036555001439719544],dtype=np.float64)

  for i in numba.prange(n):
    hmag = np.float64(0.0)
    for j in range(3):
      hmag += ho[i,j] * ho[i,j]
    if hmag < eps:
      ax[i,0] = 0.0
      ax[i,1] = 0.0
      ax[i,2] = -1.0 * pf
      ax[i,3] = 0.0
    else:
      hm = hmag
      sqrthm = np.sqrt(hm)
      hn = np.zeros(3,dtype=intype)
      for j in range(3):
        hn[j] = ho[i,j] / sqrthm
      # hn = ho[i,:]/sqrthm
      s = tfit[0] + tfit[1] * hmag
      for j in range(2,16):
        hm *= hmag
        s += tfit[j] * hm
      s = 1.0 if (s > 1.0) else s
      s = -1.0 if (s < -1.0) else s
      s = 2.0 * np.arccos(s)

      for j in range(3):
        ax[i,j] = hn[j]
      if np.abs(s - PI) < eps:
        ax[i,3] = PI
      else:
        ax[i,3] = s

  return ax


def ro2ax(ro,p=P):
  '''Rodrigues-Frank vector to axis-angle (radians)  '''
  ax = ro2axL(np.require(ro,requirements=['C','A']),p=p)
  ax = np.squeeze(ax)
  return ax


@numba.jit(nopython=True,fastmath=nbFastmath,cache=nbcache,parallel=nbParallel)
def ro2axL(roIn,p=P):
  pf = numba.float32(p > 0) * 2.0 - 1.0
  ro,m,n,intype = prepIn(roIn)
  ax = np.zeros((n,4),dtype=intype)
  for i in numba.prange(n):
    ta = ro[i,3]
    if ta < eps:
      ax[i,0] = 0.0
      ax[i,1] = 0.0
      ax[i,2] = -1 * pf
      ax[i,3] = 0.0
    else:
      if ta == np.inf:
        ax[i,0] = ro[i,0]
        ax[i,1] = ro[i,1]
        ax[i,2] = ro[i,2]
        ax[i,3] = PI
      else:
        ax[i,3] = 2.0 * np.arctan(ta)
        ta = 0.0
        for j in range(3):
          ta += ro[i,j] * ro[i,j]
        ta = 1.0 / np.sqrt(ta)
        for j in range(3):
          ax[i,j] = ro[i,j] * ta
  return ax


def ax2ro(ax,p=P):
  '''axis-angle (radians) to Rodrigues-Frank vector  '''
  ro = ax2roL(np.require(ax,requirements=['C','A']),p=p)
  ro = np.squeeze(ro)
  return ro


@numba.jit(nopython=True,fastmath=nbFastmath,cache=nbcache,parallel=nbParallel)
def ax2roL(axIn,p=P):
  pf = numba.float32(p > 0) * 2.0 - 1.0
  ax,m,n,intype = prepIn(axIn)
  ro = np.zeros((n,4),dtype=intype)
  axn = axnormL(ax)
  for i in numba.prange(n):
    if axn[i,3] == 0:
      ro[i,0] = 0.0
      ro[i,1] = 0.0
      ro[i,2] = -1.0 * pf
      ro[i,3] = 0.0
    else:
      for j in range(3):
        ro[i,j] = axn[i,j]
      if (PI - ax[i,3]) < eps:
        ro[i,3] = np.inf
      else:
        ro[i,3] = np.tan(axn[i,3] * 0.5)
  return ro


def ax2qu(ax,p=P):
  '''axis-angle (radians) to quaternion (scalar - vector)  '''
  qu = ax2quL(np.require(ax,requirements=['C','A']),p=p)
  qu = np.squeeze(qu)
  return qu


@numba.jit(nopython=True,fastmath=nbFastmath,cache=nbcache,parallel=nbParallel)
def ax2quL(axIn,p=P):
  pf = numba.float32(p > 0) * 2.0 - 1.0
  ax,m,n,intype = prepIn(axIn)
  qu = np.zeros((n,4),dtype=intype)
  axn = axnormL(ax)
  #axn = ax
  for i in numba.prange(n):
    if axn[i,3] < eps:
      qu[i,0] = 1.0
      qu[i,1] = 0.0
      qu[i,2] = 0.0
      qu[i,3] = 0.0
    else:
      qu[i,0] = np.cos(ax[i,3] * 0.5)
      mag = 0.0
      for j in range(3):
        mag += ax[i,j] * ax[i,j]
      s = np.sin(ax[i,3] * 0.5) / mag
      for j in range(3):
        qu[i,j + 1] = ax[i,j] * s
  qu = quatnormL(qu)
  return qu


def ro2ho(ro,p=P):
  '''Rodrigues-Frank  to homochoric  '''
  ho = ro2hoL(np.require(ro,requirements=['C','A']),p=p)
  ho = np.squeeze(ho)
  return ho


@numba.jit(nopython=True,fastmath=nbFastmath,cache=nbcache,parallel=nbParallel)
def ro2hoL(roIn,p=P):
  pf = numba.float32(p > 0) * 2.0 - 1.0
  ro,m,n,intype = prepIn(roIn)
  ho = np.zeros((n,3),dtype=intype)
  for i in numba.prange(n):
    mag = ro[i,3]
    if mag < eps:
      ho[i,0] = 0.0
      ho[i,1] = 0.0
      ho[i,2] = 0.0
    else:
      if mag > 1e9:
        f = 0.75 * PI
      else:
        t = 2.0 * np.arctan(ro[i,3])
        f = 0.75 * (t - np.sin(t))
      f = f ** (1.0 / 3.0)
      for j in range(3):
        ho[i,j] = ro[i,j] * f

  return ho


def qu2om(qu,p=P):
  '''quaternion (scalar-vector) to orientation matrix  '''
  om = qu2omL(np.require(qu,requirements=['C','A']),p=p)
  om = np.squeeze(om)
  return om


@numba.jit(nopython=True,fastmath=nbFastmath,cache=nbcache,parallel=nbParallel)
def qu2omL(quIn,p=P):
  pf = numba.float32(p > 0) * 2.0 - 1.0
  qu,m,n,intype = prepIn(quIn)
  om = np.zeros((n,3,3),dtype=intype)
  # [0,0], [0,1], [0, 2], [1,0], [1,1], [1,2], [2,0], [2,1], [2,2]
  #    0      1      2      3      4      5      6      7      8
  qu = quatnormL(qu)


  for i in numba.prange(n):
    q0 = qu[i,0] * pf
    qq = q0 * q0 - (qu[i,1] ** 2 + qu[i,2] ** 2 + qu[i,3] ** 2)
    om[i,0,0] = qq + 2.0 * qu[i,1] * qu[i,1]
    om[i,0,1] = 2.0 * (qu[i,1] * qu[i,2] - qu[i,0] * qu[i,3])
    om[i,0,2] = 2.0 * (qu[i,1] * qu[i,3] + qu[i,0] * qu[i,2])
    om[i,1,0] = 2.0 * (qu[i,2] * qu[i,1] + qu[i,0] * qu[i,3])
    om[i,1,1] = qq + 2.0 * qu[i,2] * qu[i,2]
    om[i,1,2] = 2.0 * (qu[i,2] * qu[i,3] - qu[i,0] * qu[i,1])
    om[i,2,0] = 2.0 * (qu[i,3] * qu[i,1] - qu[i,0] * qu[i,2])
    om[i,2,1] = 2.0 * (qu[i,3] * qu[i,2] + qu[i,0] * qu[i,1])
    om[i,2,2] = qq + 2.0 * qu[i,3] * qu[i,3]

    tr = om[i,0,0] + om[i,1,1] + om[i,2,2]
    t = 0.50 * (tr - 1.0)
    if np.abs(t) > 1.0:  # there has been enough rounding to justify a re normalization of the matrix
      #using a Gram-Schmidt normalization
      u1 = om[i,0,:]
      u1 /= np.linalg.norm(u1)
      om[i,0,:] = u1
      u2 = om[i,1,:]
      u2 = u2-np.sum(u2*u1)*u1
      u2 /= np.linalg.norm(u2)
      om[i,1,:] = u2
      u3 = om[i,2,:]
      u3 - np.sum(u3*u1)*u1 - np.sum(u3*u2)*u2
      u3 /= np.linalg.norm(u3)
      om[i,2,:] = u3

  return om

def om2ax_old(om,p=P):
  '''orientation matrix to axis-angle (radians) -- depreciated now use qu2ax(om2qu()) '''
  if om.ndim == 3:
    omIn = np.require(om,requirements=['C','A'])
  elif om.ndim == 2:
    omIn = np.require(om,requirements=['C','A']).reshape((1,3,3))
  else:
    return -1
  ax = om2axL(omIn,p=p)
  ax = np.squeeze(ax)
  return ax


@numba.jit(nopython=True,fastmath=nbFastmath, cache=nbcache, parallel=nbParallel)
def om2axL(om, p=P):# depreciated now use qu2ax(om2qu()) -- kept for historical reasons.
  pf = numba.float32(p > 0) * 2.0 - 1.0
  intype = om.dtype
  n = np.int64(om.size / 9)
  ax = np.zeros((n, 4), dtype=intype)
  #  help for translating out of C version
  # [0,0], [0,1], [0, 2], [1,0], [1,1], [1,2], [2,0], [2,1], [2,2] 
  #    0      1      2      3      4      5      6      7      8
  for i in numba.prange(n):

    tr = om[i,0,0] + om[i,1,1] + om[i,2,2]
    t = 0.50 * (tr - 1.0)
    t = 1.0 if (t > 1.0) else t
    t = -1.0 if (t < -1.0) else t
    ax[i,3] = np.arccos(t)
    #if ((1.0 - np.abs(t)) > eps):
    mag = numba.float64(0.0)
    ax[i,2] = pf * (om[i,1,0] - om[i,0,1])
    mag += ax[i,2]*ax[i,2]
    ax[i,1] = pf * (om[i,0,2] - om[i,2,0])
    mag += ax[i,1] * ax[i,1]
    ax[i,0] = pf * (om[i,2,1] - om[i,1,2])
    mag += ax[i,0] * ax[i,0]
    mag = np.sqrt(mag)
    if mag > eps:
      for j in range(3):
        ax[i,j] *= 1.0/mag
    else:
      if t > 0.0:
        ax[i, 0] = 0.0
        ax[i, 1] = 0.0
        ax[i, 2] = -1.0*pf
      else:
        d = np.zeros(3,dtype=intype)
        for j in range(3):
          d[j] = np.sqrt(0.5*(om[i,j,j]+1.0))
        dargsrt = np.argsort(d)
        d[dargsrt[1]] = (om[i, dargsrt[2], dargsrt[1]] + om[i, dargsrt[1], dargsrt[2] ]) / (4.0 * d[dargsrt[2]])
        d[dargsrt[0]] = (om[i, dargsrt[2], dargsrt[0]] + om[i, dargsrt[0], dargsrt[2]]) / (4.0 * d[dargsrt[2]])

        for j in range(3):
          ax[i,j] = pf*d[j]

  ax = axnormL(ax)
  return ax

@numba.jit(nopython=True,fastmath=False, cache=nbcache, parallel=nbParallel)
def om2axL_2(om, p=P): # kept for historical reasons
  # ** not sure it gets the axis correct when angle = pi **
  pf = numba.float32(p > 0) * 2.0 - 1.0
  intype = om.dtype
  n = np.int64(om.size / 9)
  ax = np.zeros((n, 4), dtype=intype)
  #  help for translating out of C version
  # [0,0], [0,1], [0, 2], [1,0], [1,1], [1,2], [2,0], [2,1], [2,2]
  #    0      1      2      3      4      5      6      7      8
  for i in numba.prange(n):

    tr = om[i,0,0] + om[i,1,1] + om[i,2,2]
    t = 0.50 * (tr - 1.0)
    t = 1.0 if (t > 1.0) else t
    t = -1.0 if (t < -1.0) else t
    ax[i,3] = np.arccos(t)
    if (ax[i,3] < 1.0e-7):
      ax[i,0] = 0.0
      ax[i,1] = 0.0
      ax[i,2] = -1.0 * pf
    else:
      omcomplex = np.zeros((3,3), dtype=numba.complex128)
      omcomplex[:,:] = om[i,:,:]
      eigval, eigvect = np.linalg.eig(omcomplex)
      whmin = np.argmin(np.absolute(eigval-1.0))
      vect = np.real(eigvect[:,whmin])
      d = (om[i,2,1] - om[i,1,2])
      if d != 0.0:
        vect[0] = np.abs(vect[0])*np.sign(d)

      d = (om[i,0,2] - om[i,2,0])
      if d != 0.0:
        vect[1] = np.abs(vect[1]) * np.sign(d)

      d = (om[i,1,0] - om[i,0,1])
      if d != 0.0:
        vect[2] = np.abs(vect[2]) * np.sign(d)

      ax[i,0] = pf*vect[0]
      ax[i,1] = pf*vect[1]
      ax[i,2] = pf*vect[2]
  ax = axnormL(ax)
  return ax

def om2qu(om,p=P):
  '''orientation matrix to quaternion (scalar-vector)  '''
  if om.ndim == 3:
    omIn = np.require(om,requirements=['C','A'])
  elif om.ndim == 2:
    omIn = np.require(om,requirements=['C','A']).reshape((1,3,3))
  else:
    return -1
  qu = om2quL(omIn,p=p)
  qu = np.squeeze(qu)
  return qu


@numba.jit(nopython=True,fastmath=nbFastmath, cache=nbcache, parallel=nbParallel)
def om2quL(om, p=P):
  """
      Thanks to Martin-Diehl using the same formulation as his rotations.py package.
      This fixes the ambiguity of the axis direction when the angle = pi.
      This also means that om2ax is now qu2ax(om2qu())
      This formulation is from  www.euclideanspace.com/maths/geometry/rotations/conversions/matrixToQuaternion.
      The original formulation had issues.
  """
  pf = numba.float32(p > 0) * 2.0 - 1.0
  intype = om.dtype
  n = np.int64(om.size / 9)
  qu = np.zeros((n, 4), dtype=intype)

  for i in numba.prange(n):
    s = 0.0
    tr = om[i,0,0] + om[i,1,1] + om[i,2,2]
    if tr > 0:
      s = 0.5 / np.sqrt(tr + 1.0)
      qu[i,0] = 0.25/s
      qu[i,1] = s * (om[i,2,1] - om[i,1,2])
      qu[i,2] = s * (om[i,0,2] - om[i,2,0])
      qu[i,3] = s * (om[i,1,0] - om[i,0,1])
    else:
      if (om[i,0,0] > om[i,1,1]) and (om[i,0,0] > om[i,2,2]):
        s = 2.0 * np.sqrt( 1.0 + om[i,0,0] - om[i,1,1] - om[i,2,2])
        qu[i,0] = (om[i,2,1] - om[i,1,2]) / s
        qu[i,1] = 0.25 * s
        qu[i,2] = (om[i,0,1] + om[i,1,0]) / s
        qu[i,3] = (om[i,0,2] + om[i,2,0]) / s
      elif om[i,1,1] > om[i,2,2]:
        s = 2.0 * np.sqrt(1.0 - om[i,0,0] + om[i,1,1] - om[i,2,2])
        qu[i,0] = (om[i,0,2] - om[i,2,0]) / s
        qu[i,1] = (om[i,0,1] + om[i,1,0]) / s
        qu[i,2] = 0.25 * s
        qu[i,3] = (om[i,1,2] + om[i,2,1]) / s
      else:
        s = 2.0 * np.sqrt(1.0 - om[i,0,0] - om[i,1,1] + om[i,2,2])
        qu[i,0] = (om[i,1,0] - om[i,0,1]) / s
        qu[i,1] = (om[i,0,2] + om[i,2,0]) / s
        qu[i,2] = (om[i,1,2] + om[i,2,1]) / s
        qu[i,3] = 0.25 * s
  qu = quatnormL(qu)
  return qu

def qu2ax(qu,p=P):
  '''quaternion (scalar-vector) to axis-angle (radians)  '''
  ax = qu2axL(np.require(qu,requirements=['C','A']),p=p)
  ax = np.squeeze(ax)
  return ax


@numba.jit(nopython=True,fastmath=nbFastmath,cache=nbcache,parallel=nbParallel)
def qu2axL(quIn,p=P):
  pf = numba.float32(p > 0) * 2.0 - 1.0
  qu,m,n,intype = prepIn(quIn)
  qu = quatnormL(qu)
  ax = np.zeros((n,4),dtype=intype)

  for i in numba.prange(n):

    if qu[i,0] > (1.0-eps):
      ax[i,0] = 0.0
      ax[i,1] = 0.0
      ax[i,2] = -pf
      ax[i,3] = 0.0
    else:
      omega = 2.0 * np.arccos(qu[i,0])
      mag = 0.0
      for j in range(3):
        mag += qu[i,j + 1] ** 2
      mag = 1.0 / (np.sqrt(mag)+1e-35)
      for j in range(3):
        ax[i,j] = qu[i,j + 1] * mag
      ax[i,3] = omega

  return ax


def qu2ro(qu,p=P):
  '''quaternion (scalar-vector) to Rodrigues-Frank vector  '''
  ro = qu2roL(np.require(qu,requirements=['C','A']),p=p)
  ro = np.squeeze(ro)
  return ro


@numba.jit(nopython=True,fastmath=nbFastmath,cache=nbcache,parallel=nbParallel)
def qu2roL(quIn,p=P):
  pf = numba.float32(p > 0) * 2.0 - 1.0
  qu,m,n,intype = prepIn(quIn)
  qu = quatnormL(qu)
  ro = np.zeros((n,4),dtype=intype)

  for i in numba.prange(n):

    if qu[i,0] < eps:
      ro[i,3] = np.inf
      for j in range(3):
        ro[i,j] = qu[i,j + 1]
    else:
      mag = 0.0
      for j in range(3):
        mag += qu[i,j + 1] ** 2
      if mag < eps:
        ro[0] = 0.0
        ro[1] = 0.0
        ro[2] = -pf
        ro[3] = 0.0
      else:
        mag = 1.0 / np.sqrt(mag)
        q0 = qu[i,0]
        q0 = 1.0-eps if q0 > 1.0 else q0
        q0 = -1.0+eps if q0 < -1.0 else q0
        ro[i,3] = np.tan(np.arccos(q0))
        for j in range(3):
          ro[i,j] = qu[i,j + 1] * mag
  return ro


def qu2ho(qu,p=P):
  '''quaternion (scalar-vector) to homochoric  '''
  ho = qu2hoL(np.require(qu,requirements=['C','A']),p=p)
  ho = np.squeeze(ho)
  return ho


@numba.jit(nopython=True,fastmath=nbFastmath,cache=nbcache,parallel=nbParallel)
def qu2hoL(quIn,p=P):
  pf = numba.float32(p > 0) * 2.0 - 1.0
  qu,m,n,intype = prepIn(quIn)
  qu = quatnormL(qu)
  ho = np.zeros((n,3),dtype=intype)

  for i in numba.prange(n):

    if qu[i,0] > (1.0-eps):
      ho[i,0] = 0.0
      ho[i,1] = 0.0
      ho[i,2] = 0.0
    else:
      omega = 2.0 * np.arccos(qu[i,0])
      mag = 0.0
      for j in range(3):
        mag += qu[i,j + 1] ** 2
      f = 0.75 * (omega - np.sin(omega))
      f = f ** (1.0 / 3.0)
      mag = f / np.sqrt(mag)
      for j in range(3):
        ho[i,j] = qu[i,j + 1] * mag
  return ho

def ho2cu(ho,p=P):
  ''' homochoric to cubochoric  '''
  cu = ho2cuL(np.require(ho,requirements=['C','A']),p=p)
  cu = np.squeeze(cu)
  return cu


@numba.jit(nopython=True,fastmath=nbFastmath,cache=nbcache, parallel=nbParallel)
def ho2cuL(hoIn,p=P):
  pf = numba.float32(p > 0) * 2.0 - 1.0
  ho,m,n,intype = prepIn(hoIn)
  cu = np.zeros((n,3),dtype=intype)
  LPR1 = np.float(1.33067003949147)  # (3pi/4)**(1/3)
  LPpref = np.float(1.38197659788534)  # sqrt(6/pi)
  LPbeta = np.float(0.962874509979126)  # pi**(5/6)/6**(1/6)/2
  LPr2 = np.float(1.4142135623731)  # sqrt(2)
  LPpi12 = np.float(0.261799387799149)  # pi/12
  LPsc = np.float(0.897772786961286)  # a/ap == (pi**(5/6)/6**(1/6)) / pi**(2/3)
  eps_loc = 1e-7



  for i in numba.prange(n):
    #cu[i,:] = lambert3DBallToCube(ho[i,:])
    # I have inlined all the function of lambert3DCubeToBall
    # numba really dislikes nested function calls.  This was about a 10x speedup.
    xyz = np.zeros(3,dtype=np.float64)
    xyzcu = np.zeros(3,dtype=np.float64)
    xyz3 = np.zeros(3,dtype=np.float64)
    xyz2 = np.zeros(3,dtype=np.float64)
    xyz1 = np.zeros(3,dtype=np.float64)
    rs = 0.0
    for j in range(3):
      xyz[j] = ho[i,j]
      rs += xyz[j] * xyz[j]
    rs = np.sqrt(np.float64(max(rs,0.0)))

    # if rs > (LPR1 + eps_loc):
    #  return xyzcu-np.float64(5.0) # out of range
    py = 10
    if (np.abs(xyz[0]) <= xyz[2]) and (np.abs(xyz[1]) <= xyz[2]):
      py = 1
    if (np.abs(xyz[0]) <= -xyz[2]) and (np.abs(xyz[1]) <= -xyz[2]):
      py = 2
    if (np.abs(xyz[2]) <= xyz[0]) and (np.abs(xyz[1]) <= xyz[0]):
      py = 3
    if (np.abs(xyz[2]) <= -xyz[0]) and (np.abs(xyz[1]) <= -xyz[0]):
      py = 4
    if (np.abs(xyz[0]) <= xyz[1]) and (np.abs(xyz[2]) <= xyz[1]):
      py = 5
    if (np.abs(xyz[0]) <= -xyz[1]) and (np.abs(xyz[2]) <= -xyz[1]):
      py = 6

    if (py == 1) or (py == 2):
      xyz3[0] = xyz[0]
      xyz3[1] = xyz[1]
      xyz3[2] = xyz[2]
    if (py == 3) or (py == 4):
      xyz3[0] = xyz[1]
      xyz3[1] = xyz[2]
      xyz3[2] = xyz[0]
    if (py == 5) or (py == 6):
      xyz3[0] = xyz[2]
      xyz3[1] = xyz[0]
      xyz3[2] = xyz[1]

    q = np.sqrt(np.float64(max(2.0 * rs / (rs + np.abs(xyz3[2]) + eps),0.0)))
    xyz2[0] = xyz3[0] * q
    xyz2[1] = xyz3[1] * q
    sign = 1.0 if (xyz3[2] >= 0.0) else -1.0
    xyz2[2] = sign * rs / LPpref

    qxy = 0.0
    for j in range(2):
      qxy += xyz2[j] * xyz2[j]

    if qxy == 0.0:
      xyz1[0] = 0.0
      xyz1[1] = 0.0
    else:
      xyswitch = np.abs(xyz2[1]) <= np.abs(xyz2[0])
      if xyswitch:
        x = xyz2[0]
        y = xyz2[1]
      else:
        x = xyz2[1]
        y = xyz2[0]
      q2xy = qxy + x * x
      sq2xy = np.sqrt(q2xy)

      q = q2xy * qxy / (q2xy - np.abs(x) * sq2xy)
      q = np.sqrt(np.float64(max(q,0.0)))
      q *= LPbeta / LPr2 / LPR1

      tt = (y * y + np.abs(x) * sq2xy) / LPr2 / qxy
      if tt > 1.0:
        tt = 1.0
      if tt < -1.0:
        tt = -1.0

      ac = np.arccos(tt)
      sx = 1.0 if (x >= 0.0) else -1.0
      sy = 1.0 if (y >= 0.0) else -1.0

      T1inv = q * sx
      T2inv = q * sy * ac / LPpi12

      if xyswitch:
        xyz1[0] = T1inv
        xyz1[1] = T2inv
      else:
        xyz1[1] = T1inv
        xyz1[0] = T2inv

    xyz1[2] = xyz2[2]
    for j in range(3):
      xyz1[j] *= 1.0 / LPsc

    if (py == 1) or (py == 2):
      xyzcu[0] = xyz1[0]
      xyzcu[1] = xyz1[1]
      xyzcu[2] = xyz1[2]
    if (py == 3) or (py == 4):
      xyzcu[0] = xyz1[2]
      xyzcu[1] = xyz1[0]
      xyzcu[2] = xyz1[1]
    if (py == 5) or (py == 6):
      xyzcu[0] = xyz1[1]
      xyzcu[1] = xyz1[2]
      xyzcu[2] = xyz1[0]
    for j in range(3):
      cu[i, j] = xyzcu[j]
  return cu

def cu2ho(cu,p=P):
  ''' cubochoric to homochoric  '''
  ho = cu2hoL(np.require(cu,requirements=['C','A']),p=p)
  ho = np.squeeze(ho)
  return ho


@numba.jit(nopython=True,fastmath=nbFastmath,cache=nbcache, parallel=nbParallel)
def cu2hoL(cuIn,p=P):
  pf = numba.float32(p > 0) * 2.0 - 1.0
  cu,m,n,intype = prepIn(cuIn)
  ho = np.zeros((n,3),dtype=intype)
  LPap = np.float64(2.14502939711103)  # pi**(2/3)
  LPa = np.float64(1.92574901995825)  # pi**(5/6)/6**(1/6)
  LPsc = np.float64(0.897772786961286)  # a/ap == (pi**(5/6)/6**(1/6)) / pi**(2/3)
  LPpref = np.float64(1.38197659788534)  # sqrt(6/pi)
  LPpi12 = np.float64(0.261799387799149)  # pi/12
  LPr2 = np.float64(1.4142135623731)  # sqrt(2)
  LPprek = np.float64(1.6434564029725)  # R1 2**(1/4)/beta == (3pi/4)^(1/3) * 2**(1/4) / (pi**(5/6)/6**(1/6)/2)
  LPsPi = np.float64(1.77245385090552)  # sqrt(pi)
  LPr24 = np.float64(4.89897948556636)  # sqrt(24)
  eps_loc = 2e-6



  # I have inlined all the function of lambert3DCubeToBall
  # numba really dislikes nested function calls.  This was about a 10x speedup.
  # for i in numba.prange(n):
  #   ho[i, :] = lambert3DCubeToBall(cu[i,:])

  for i in numba.prange(n):
    xyz = np.zeros(3,dtype=np.float64)
    xyzba = np.zeros(3,dtype=np.float64)
    sXYZ = np.zeros(3,dtype=np.float64)
    #xyz[:] = cu[i,:]
    xyz[0] = cu[i,0]
    xyz[1] = cu[i,1]
    xyz[2] = cu[i,2]
    mx = np.abs(xyz).max()

    # if mx > (LPap/2.0 + eps_loc):
    #  return np.float64([-5,-5, -5]) #xyzba out of range

    py = 10
    if (np.abs(xyz[0]) <= xyz[2]) and (np.abs(xyz[1]) <= xyz[2]):
      py = 1
    if (np.abs(xyz[0]) <= -xyz[2]) and (np.abs(xyz[1]) <= -xyz[2]):
      py = 2
    if (np.abs(xyz[2]) <= xyz[0]) and (np.abs(xyz[1]) <= xyz[0]):
      py = 3
    if (np.abs(xyz[2]) <= -xyz[0]) and (np.abs(xyz[1]) <= -xyz[0]):
      py = 4
    if (np.abs(xyz[0]) <= xyz[1]) and (np.abs(xyz[2]) <= xyz[1]):
      py = 5
    if (np.abs(xyz[0]) <= -xyz[1]) and (np.abs(xyz[2]) <= -xyz[1]):
      py = 6

    if (py == 1) or (py == 2):
      sXYZ[0] = xyz[0]
      sXYZ[1] = xyz[1]
      sXYZ[2] = xyz[2]
    if (py == 3) or (py == 4):
      sXYZ[0] = xyz[1]
      sXYZ[1] = xyz[2]
      sXYZ[2] = xyz[0]
    if (py == 5) or (py == 6):
      sXYZ[0] = xyz[2]
      sXYZ[1] = xyz[0]
      sXYZ[2] = xyz[1]

    for j in range(3):
      sXYZ[j] = sXYZ[j] * LPsc

    if mx < eps:
      sXYZ[:] = 0.0
    else:
      absX = np.abs(sXYZ[0])
      absY = np.abs(sXYZ[1])

      if (absX < eps) and (absY < eps):
        sXYZ[0] = 0.0
        sXYZ[1] = 0.0
        sXYZ[2] = LPpref * sXYZ[2]
      else:
        z = np.float64(sXYZ[2])
        xyswitch = np.abs(sXYZ[1]) <= np.abs(sXYZ[0])
        if xyswitch:
          x = np.float64(sXYZ[0])
          y = np.float64(sXYZ[1])
        else:
          x = np.float64(sXYZ[1])
          y = np.float64(sXYZ[0])

        q = LPpi12 * y / x
        c = np.float64(np.cos(q))
        s = np.float64(np.sin(q))

        q = LPprek * x / np.sqrt(np.float64(max(LPr2 - c,0.0)))
        T1p = (LPr2 * c - 1.0) * q
        T2p = LPr2 * s * q
        if xyswitch:
          T1 = np.float64(T1p)
          T2 = np.float64(T2p)
        else:
          T1 = np.float64(T2p)
          T2 = np.float64(T1p)

        c = T1 * T1 + T2 * T2
        s = np.pi * c / (24.0 * z * z)

        c = LPsPi * c / LPr24 / z
        q = np.sqrt(np.float64(max([1.0 - s,0.0])))

        sXYZ[0] = T1 * q
        sXYZ[1] = T2 * q
        sXYZ[2] = LPpref * z - c

    if (py == 1) or (py == 2):
      xyzba[0] = sXYZ[0]
      xyzba[1] = sXYZ[1]
      xyzba[2] = sXYZ[2]
    if (py == 3) or (py == 4):
      xyzba[0] = sXYZ[2]
      xyzba[1] = sXYZ[0]
      xyzba[2] = sXYZ[1]
    if (py == 5) or (py == 6):
      xyzba[0] = sXYZ[1]
      xyzba[1] = sXYZ[2]
      xyzba[2] = sXYZ[0]

    for j in range(3):
      ho[i,j] = xyzba[j]

  return ho

# cross dependant codes start here
def eu2ro(eu,p=P):
  '''Bunge Euler angles to Rodrigues-Frank vector  '''
  ro = ax2ro(eu2ax(eu, p=p), p=p)
  return ro

def ro2om(ro,p=P):
  '''Rodrigues-Frank vector to orientation matrix   '''
  om = ax2om(ro2ax(ro, p=p), p=p)
  return om

def ro2eu(ro,p=P):
  '''Rodrigues-Frank vector to Bunge Euler angles  '''
  eu = om2eu(ro2om(ro, p=p), p=p)
  return eu

def eu2ho(eu,p=P):
  '''Bunge Euler angles to homochoric  '''
  ho = ax2ho(eu2ax(eu, p=p), p=p)
  return ho

def om2qu_old(om,p=P):
  '''orientation matrix to quaternion (scalar-vector)  '''
  qu = ax2qu(om2ax(om, p=p), p=p)
  return qu

def om2ax(om,p=P):
  '''orientation matrix to quaternion (scalar-vector)  '''
  ax = qu2ax(om2qu(om, p=p), p=p)
  return ax


def om2ro(om,p=P):
  '''orientation matrix to Rodrigues-Frank vector  '''
  ro = eu2ro(om2eu(om, p=p), p=p)
  return ro

def om2ho(om,p=P):
  '''orientation matrix to homochoric  '''
  ho = ax2ho(om2ax(om, p=p), p=p)
  return ho

def ax2eu(ax,p=P):
  '''axis-angle to Bunge Euler angle  '''
  eu = om2eu(ax2om(ax, p=p), p=p)
  return eu

def ro2qu(ro,p=P):
  '''Rodrigues-Frank vector to quaternion (scalar-vector)  '''
  qu = ax2qu(ro2ax(ro, p=p), p=p)
  return qu

def ho2eu(ho,p=P):
  '''homochoric to Bunge Euler angles  '''
  eu = ax2eu(ho2ax(ho, p=p), p=p)
  return eu

def ho2om(ho,p=P):
  '''homochoric to orientation matrix  '''
  om = ax2om(ho2ax(ho, p=p), p=p)
  return om

def ho2ro(ho,p=P):
  '''homochoric to Rodrigues-Frank vector  '''
  ro = ax2ro(ho2ax(ho, p=p), p=p)
  return ro

def ho2qu(ho,p=P):
  '''homochoric to quaternion (scalar-vector)  '''
  qu = ax2qu(ho2ax(ho, p=p), p=p)
  return qu

def eu2cu(eu,p=P):
  '''Bunge Euler angles in cubochoric  '''
  cu = ho2cu(eu2ho(eu, p=p), p=p)
  return cu

def om2cu(om,p=P):
  '''orientation matrix to cubochoric  '''
  cu = ho2cu(om2ho(om, p=p), p=p)
  return cu

def ax2cu(ax,p=P):
  '''axis-angle to cubochoric  '''
  cu = ho2cu(ax2ho(ax, p=p), p=p)
  return cu

def ro2cu(ro,p=P):
  '''Rodrigues-Frank vector to cubochoric  '''
  cu = ho2cu(ro2ho(ro, p=p), p=p)
  return cu

def qu2cu(qu,p=P):
  '''quaternion (scalar-vector) to cubochoric  '''
  cu = ho2cu(qu2ho(qu, p=p), p=p)
  return cu

def cu2eu(cu,p=P):
  '''cubochoric to Bunge Euler angles  '''
  eu = ho2eu(cu2ho(cu, p=p), p=p)
  return eu

def cu2om(cu,p=P):
  '''cubochoric to orientation matrix  '''
  om = ho2om(cu2ho(cu, p=p), p=p)
  return om

def cu2ax(cu,p=P):
  '''cubochoric to axis-angle  '''
  ax = ho2ax(cu2ho(cu, p=p), p=p)
  return ax

def cu2ro(cu,p=P):
  '''cubochoric to Rodrigues-Frank vector  '''
  ro = ho2ro(cu2ho(cu, p=p), p=p)
  return ro

def cu2qu(cu,p=P):
  '''cubochoric to quaternion (scalar-vector)  '''
  qu = ho2qu(cu2ho(cu, p=p), p=p)
  return qu

# Lambert Ball <--> Cube codes


@numba.jit(['int32(f8[:])','int32(f4[:])'], nopython=True,fastmath=nbFastmath,cache=nbcache)
def lambertGetPyramid(xyz):
  out = numba.int32(10)
  if (np.abs(xyz[0]) <= xyz[2]) and (np.abs(xyz[1]) <= xyz[2]):
    out = 1
  if (np.abs(xyz[0]) <= -xyz[2] ) and (np.abs(xyz[1]) <= -xyz[2]):
    out = 2
  if (np.abs(xyz[2]) <= xyz[0]) and (np.abs(xyz[1]) <= xyz[0]):
    out = 3
  if (np.abs(xyz[2]) <= -xyz[0]) and (np.abs(xyz[1]) <= -xyz[0]):
    out = 4
  if (np.abs(xyz[0]) <= xyz[1]) and (np.abs(xyz[2]) <= xyz[1]):
    out = 5
  if (np.abs(xyz[0]) <= -xyz[1]) and (np.abs(xyz[2]) <= -xyz[1]):
    out = 6
  return out

@numba.jit(['f8[:](f8[:])','f8[:](f4[:])'], nopython=True,fastmath=nbFastmath,cache=nbcache)
def lambert3DCubeToBall(xyz):
  LPap = np.float64(2.14502939711103) # pi**(2/3)
  LPa =   np.float64(1.92574901995825) # pi**(5/6)/6**(1/6)
  LPsc =  np.float64(0.897772786961286) # a/ap == (pi**(5/6)/6**(1/6)) / pi**(2/3)
  LPpref =  np.float64(1.38197659788534) # sqrt(6/pi)
  LPpi12 =  np.float64(0.261799387799149) # pi/12
  LPr2 =  np.float64(1.4142135623731) # sqrt(2)
  LPprek =  np.float64(1.6434564029725) # R1 2**(1/4)/beta == (3pi/4)^(1/3) * 2**(1/4) / (pi**(5/6)/6**(1/6)/2)
  LPsPi =  np.float64(1.77245385090552) # sqrt(pi)
  LPr24 =  np.float64(4.89897948556636) # sqrt(24)
  eps_loc = 2e-6

  xyzba = np.zeros(3, dtype = np.float64)
  sXYZ = np.zeros(3, dtype = np.float64)
  mx = np.abs(xyz).max()

  #if mx > (LPap/2.0 + eps_loc):
  #  return np.float64([-5,-5, -5]) #xyzba out of range

  p = lambertGetPyramid(xyz)

  if (p == 1) or (p == 2):
    sXYZ[0] = xyz[0]
    sXYZ[1] = xyz[1]
    sXYZ[2] = xyz[2]
  if (p == 3) or (p == 4):
    sXYZ[0] = xyz[1]
    sXYZ[1] = xyz[2]
    sXYZ[2] = xyz[0]
  if (p == 5) or (p == 6):
    sXYZ[0] = xyz[2]
    sXYZ[1] = xyz[0]
    sXYZ[2] = xyz[1]

  for i in range(3):
    sXYZ[i] = sXYZ[i] * LPsc


  if mx < eps:
    sXYZ[:] = 0.0
  else:
    absX = np.abs(sXYZ[0])
    absY = np.abs(sXYZ[1])


    if (absX < eps) and (absY < eps):
      sXYZ[0] = 0.0
      sXYZ[1] = 0.0
      sXYZ[2] = LPpref *sXYZ[2]
    else:
      z = np.float64(sXYZ[2])
      xyswitch = np.abs(sXYZ[1]) <= np.abs(sXYZ[0])
      if xyswitch:
        x = np.float64(sXYZ[0])
        y = np.float64(sXYZ[1])
      else:
        x = np.float64(sXYZ[1])
        y = np.float64(sXYZ[0])

      q = LPpi12 * y / x
      c = np.float64(np.cos(q))
      s = np.float64(np.sin(q))

      q = LPprek * x/np.sqrt(np.float64(max(LPr2 - c, 0.0)))
      T1p = (LPr2 * c - 1.0) * q
      T2p = LPr2 *  s * q
      if xyswitch:
        T1 = np.float64(T1p)
        T2 = np.float64(T2p)
      else:
        T1 = np.float64(T2p)
        T2 = np.float64(T1p)

      c = T1*T1 + T2*T2
      s = np.pi * c / (24.0 * z *z)

      c = LPsPi * c / LPr24 / z
      q = np.sqrt(np.float64(max([1.0-s, 0.0])))

      sXYZ[0] = T1 * q
      sXYZ[1] = T2 * q
      sXYZ[2] = LPpref * z -c

  if (p == 1) or (p == 2):
    xyzba[0] = sXYZ[0]
    xyzba[1] = sXYZ[1]
    xyzba[2] = sXYZ[2]
  if (p == 3) or (p == 4):
    xyzba[0] = sXYZ[2]
    xyzba[1] = sXYZ[0]
    xyzba[2] = sXYZ[1]
  if (p == 5) or (p == 6):
    xyzba[0] = sXYZ[1]
    xyzba[1] = sXYZ[2]
    xyzba[2] = sXYZ[0]
  return xyzba


@numba.jit(['f8[:](f8[:])','f8[:](f4[:])'], nopython=True,fastmath=nbFastmath, cache=nbcache)
def lambert3DBallToCube(xyz):
  LPR1 = np.float(1.33067003949147) # (3pi/4)**(1/3)
  LPpref = np.float(1.38197659788534) # sqrt(6/pi)
  LPbeta = np.float(0.962874509979126) # pi**(5/6)/6**(1/6)/2
  LPr2 = np.float(1.4142135623731) # sqrt(2)
  LPpi12 = np.float(0.261799387799149) # pi/12
  LPsc = np.float(0.897772786961286) # a/ap == (pi**(5/6)/6**(1/6)) / pi**(2/3)
  eps_loc = 1e-7

  xyzcu = np.zeros(3, dtype = np.float64)
  xyz3 = np.zeros(3,dtype=np.float64)
  xyz2 = np.zeros(3,dtype=np.float64)
  xyz1 = np.zeros(3,dtype=np.float64)
  rs = 0.0
  for i in range(3):
    rs += xyz[i] * xyz[i]
  rs = np.sqrt(np.float64(max(rs, 0.0)))

  #if rs > (LPR1 + eps_loc):
  #  return xyzcu-np.float64(5.0) # out of range

  p = lambertGetPyramid(xyz)
  if (p == 1) or (p == 2):
    xyz3[0] = xyz[0]
    xyz3[1] = xyz[1]
    xyz3[2] = xyz[2]
  if (p == 3) or (p == 4):
    xyz3[0] = xyz[1]
    xyz3[1] = xyz[2]
    xyz3[2] = xyz[0]
  if (p == 5) or (p == 6):
    xyz3[0] = xyz[2]
    xyz3[1] = xyz[0]
    xyz3[2] = xyz[1]

  q = np.sqrt(np.float64(max(2.0 * rs/(rs+np.abs(xyz3[2])+eps), 0.0)))
  xyz2[0] = xyz3[0] * q
  xyz2[1] = xyz3[1] * q
  sign = 1.0 if (xyz3[2] >= 0.0) else -1.0
  xyz2[2] = sign * rs/LPpref

  qxy = 0.0
  for i in range(2):
    qxy += xyz2[i]*xyz2[i]

  if qxy == 0.0:
    xyz1[0] = 0.0
    xyz1[1] = 0.0
  else:
    xyswitch = np.abs(xyz2[1]) <= np.abs(xyz2[0])
    if xyswitch:
      x = xyz2[0]
      y = xyz2[1]
    else:
      x = xyz2[1]
      y = xyz2[0]
    q2xy = qxy + x * x
    sq2xy = np.sqrt(q2xy)

    q = q2xy * qxy/(q2xy - np.abs(x)*sq2xy)
    q = np.sqrt(np.float64(max(q,0.0)))
    q *= LPbeta / LPr2 / LPR1

    tt = (y * y + np.abs(x) * sq2xy) / LPr2 / qxy
    if tt > 1.0:
      tt = 1.0
    if tt < -1.0:
      tt = -1.0

    ac = np.arccos(tt)
    sx = 1.0 if (x >= 0.0) else -1.0
    sy = 1.0 if (y >= 0.0) else -1.0

    T1inv = q * sx
    T2inv = q * sy * ac/LPpi12

    if xyswitch:
      xyz1[0] = T1inv
      xyz1[1] = T2inv
    else:
      xyz1[1] = T1inv
      xyz1[0] = T2inv

  xyz1[2] = xyz2[2]
  for i in range(3):
    xyz1[i] *= 1.0/LPsc

  if (p == 1) or (p == 2):
    xyzcu[0] = xyz1[0]
    xyzcu[1] = xyz1[1]
    xyzcu[2] = xyz1[2]
  if (p == 3) or (p == 4):
    xyzcu[0] = xyz1[2]
    xyzcu[1] = xyz1[0]
    xyzcu[2] = xyz1[1]
  if (p == 5) or (p == 6):
    xyzcu[0] = xyz1[1]
    xyzcu[1] = xyz1[2]
    xyzcu[2] = xyz1[0]
  return xyzcu

# Basic quaternion functions

def quatnorm(qu):
  '''quaternion normalization -- first term >=0 , unit quaternion length'''
  qout = quatnormL(np.require(qu,requirements=['C','A']))
  qout = np.squeeze(qout)
  return qout


@numba.jit(nopython=True,fastmath=nbFastmath,cache=nbcache,parallel=nbParallel)
def quatnormL(qu,unitNorm=True):
  qin,m,n,intype = prepIn(qu)
  qout = np.zeros((n,4),dtype=intype)
  for i in numba.prange(n):
    sign = 1.0 if (qin[i,0] >= 0.0) else -1.0
    sum = 0.0
    for j in range(4):
      qout[i,j] = qin[i,j] * sign
      sum += qout[i,j] * qout[i,j]
    sum = 1.0/(np.sqrt(sum)+1.0e-35)
    if unitNorm == True:
      for j in range(4):
        qout[i,j] *= sum
  return qout

def quatconj(qu):
  '''Provides a quaternion conjugate. For rotations quaternions, this is the inverse quaternion'''
  qout = quatconjL(np.require(qu,requirements=['C','A']))
  qout = np.squeeze(qout)
  return qout


@numba.jit(nopython=True,fastmath=nbFastmath,cache=nbcache,parallel=nbParallel)
def quatconjL(quin,unitNorm=True):
  qu,m,n,intype = prepIn(quin)
  qout = np.zeros((n,4),dtype=intype)
  for i in numba.prange(n):
    qout[i,0] = qu[i,0]
    for j in range(3):
      qout[i,j+1] = -1.0*qu[i,j+1]
  return qout

def quatave(qu):
  '''Provides a quaternion average. Assumes quaterions are closely clustered'''
  qout = quataveL(np.require(qu,requirements=['C','A']))
  qout = np.squeeze(qout)
  return qout


@numba.jit(nopython=True,fastmath=nbFastmath,cache=nbcache,parallel=nbParallel)
def quataveL(quin,unitNorm=True):
  qu,m,n,intype = prepIn(quin)
  qout = np.zeros((4),dtype=intype)
  q0 = qu[0,:]
  qout += q0
  for i in numba.prange(n-1):
    sign = q0[0]*qu[i+1,0]
    for j in range(3):
      sign += q0[j+1]*qu[i+1,j+1]
    if sign >= 0.0:
      s = 1.0
    else:
      s = -1.0
    qout += s*qu[i+1,:]
  #qout /= numba.float32(n)
  qout = quatnormL(qout)
  return qout

def quat_multiply(q1,q2,p=P):
  '''quaternion multiplication. Can be q1(Nx4) x q2(N x4) or q1(4) x q2(Nx4)'''
  q1q2 = quat_multiplyL(np.require(q1,requirements=['C','A']),np.require(q2,requirements=['C','A']),p=p)
  q1q2 = np.squeeze(q1q2)
  return q1q2


@numba.jit(nopython=True,fastmath=nbFastmath,cache=nbcache)
def quat_multiplyL(q1In,q2In,p=P):
  q1,m,n1,intype = prepIn(q1In)
  q2,m,n2,intype = prepIn(q2In)
  n = max(n1,n2)
  if n1 == n2:
    q1q2 = quat_multiplyLNN(q1,q2,n,intype,p=p)
  elif n1 < n2:
    q0 = q1[0,:]
    q1q2 = quat_multiplyL1N(q0,q2,n,intype,p=p)
  elif n2 < n1:
    q0 = q2[0,:]
    q1q2 = quat_multiplyL1N(q0,q1,n,intype,p=-p)
  #q1q2 = quatnormL(q1q2,unitNorm=False)
  return q1q2


@numba.jit(nopython=True,fastmath=nbFastmath,cache=nbcache,parallel=nbParallel)
def quat_multiplyLNN(q1,q2,n,intype,p=P):
  pf = numba.float32(p > 0) * 2.0 - 1.0
  q1q2 = np.zeros((n,4),dtype=intype)
  for i in numba.prange(n):
    q1q2[i,0] = q1[i,0] * q2[i,0] - (q1[i,1] * q2[i,1] + q1[i,2] * q2[i,2] + q1[i,3] * q2[i,3])
    q1q2[i,1] = q1[i,0] * q2[i,1] + q2[i,0] * q1[i,1] + pf * (q1[i,2] * q2[i,3] - q1[i,3] * q2[i,2])
    q1q2[i,2] = q1[i,0] * q2[i,2] + q2[i,0] * q1[i,2] + pf * (q1[i,3] * q2[i,1] - q1[i,1] * q2[i,3])
    q1q2[i,3] = q1[i,0] * q2[i,3] + q2[i,0] * q1[i,3] + pf * (q1[i,1] * q2[i,2] - q1[i,2] * q2[i,1])
  return q1q2

@numba.jit(nopython=True,fastmath=nbFastmath,cache=nbcache,parallel=nbParallel)
def quat_multiplyL1N(q1,q2,n,intype,p=P):
  pf = numba.float32(p > 0) * 2.0 - 1.0
  q1q2 = np.zeros((n,4),dtype=intype)
  for i in numba.prange(n):
    q1q2[i,0] = q1[0] * q2[i,0] - (q1[1] * q2[i,1] + q1[2] * q2[i,2] + q1[3] * q2[i,3])
    q1q2[i,1] = q1[0] * q2[i,1] + q2[i,0] * q1[1] + pf * (q1[2] * q2[i,3] - q1[3] * q2[i,2])
    q1q2[i,2] = q1[0] * q2[i,2] + q2[i,0] * q1[2] + pf * (q1[3] * q2[i,1] - q1[1] * q2[i,3])
    q1q2[i,3] = q1[0] * q2[i,3] + q2[i,0] * q1[3] + pf * (q1[1] * q2[i,2] - q1[2] * q2[i,1])
  return q1q2

# @numba.jit(nopython=True,fastmath=nbFastmath,cache=nbcache,parallel=True)
# def quatMultiplyLN1(q1,q2,n,intype, p=P):
#   q1q2 = np.zeros((n,4),dtype=intype)
#   for i in numba.prange(n):
#     q1q2[i,:] = quatMultiply1(q1[i,:],q2,p=p)
#   return q1q2

@numba.jit(nopython=True,fastmath=nbFastmath,cache=nbcache)
def quat_multiply1(q1,q2,p=P):
  pf = numba.float32(p > 0) * 2.0 - 1.0
  q1q2 = np.zeros_like(q1)
  q1q2[0] = q1[0] * q2[0] - (q1[1] * q2[1] + q1[2] * q2[2] + q1[3] * q2[3])
  q1q2[1] = q1[0] * q2[1] + q2[0] * q1[1] + pf * (q1[2] * q2[3] - q1[3] * q2[2])
  q1q2[2] = q1[0] * q2[2] + q2[0] * q1[2] + pf * (q1[3] * q2[1] - q1[1] * q2[3])
  q1q2[3] = q1[0] * q2[3] + q2[0] * q1[3] + pf * (q1[1] * q2[2] - q1[2] * q2[1])
  return q1q2


def quat_vector(q,v,p=P):
  '''quaternion vector rotation. Can be q1(Nx4) x v(Nx3) or q1(4) x v(Nx3) or q1(Nx4) x v(3)'''
  vout = quat_vectorL(np.require(q,requirements=['C','A']),np.require(v,requirements=['C','A']),p=p)
  vout = np.squeeze(vout)
  return vout

@numba.jit(nopython=True,fastmath=nbFastmath,cache=nbcache)
def quat_vectorL(qIn,vIn,p=P):
  q,mq,n1,intype = prepIn(qIn)
  v,mv,n2,intype = prepIn(vIn)
  n = max(n1,n2)
  #vout = np.zeros((n,3),dtype=intype)
  if n1 == n2:
    vout = quat_vectorLNN(q,v,n,intype,p=p)
  if n1 < n2:
    q0 = q[0,:]
    vout = quat_vectorL1N(q0,v,n,intype,p=p)
  if n2 < n1:
    v0 = v[0,:]
    vout = quat_vectorLN1(q,v0,n,intype,p=p)
  return vout

@numba.jit(nopython=True,fastmath=nbFastmath,cache=nbcache,parallel=False)
def quat_vectorLNN(q,v,n,intype,p=P):
  pf = numba.float32(p > 0) * 2.0 - 1.0
  vout = np.zeros((n,3),dtype=intype)
  vtemp = np.zeros((3),dtype=intype)
  vtemp2 = np.zeros((3),dtype=intype)
  r = np.zeros((3),dtype=intype)
  w = numba.float64(0.0)
  for i in numba.prange(n):
    w = q[i,0]
    r[0] = pf * q[i,1]
    r[1] = pf * q[i,2]
    r[2] = pf * q[i,3]

    # Wow oh wow, numba hates function calls in a loop! 100x speedup
    #vtemp = np.cross(r,v[i,:])
    #vtemp = qv_crossp(r,v[i,:])
    vtemp[0] = r[1] * v[i,2] - r[2] * v[i,1]
    vtemp[1] = r[2] * v[i,0] - r[0] * v[i,2]
    vtemp[2] = r[0] * v[i,1] - r[1] * v[i,0]

    for j in range(3):
      vtemp[j] += w * v[i,j]

    #vtemp2 = np.cross(r,vtemp)
    #vtemp2 = qv_crossp(r,vtemp)
    vtemp2[0] = r[1] * vtemp[2] - r[2] * vtemp[1]
    vtemp2[1] = r[2] * vtemp[0] - r[0] * vtemp[2]
    vtemp2[2] = r[0] * vtemp[1] - r[1] * vtemp[0]


    for j in range(3):
      vout[i,j] = 2.0 * vtemp2[j] + v[i,j]

  return vout

@numba.jit(nopython=True,fastmath=nbFastmath,cache=nbcache,parallel=False)
def quat_vectorL1N(q,v,n,intype,p=P):
  pf = numba.float32(p > 0) * 2.0 - 1.0
  vout = np.zeros((n,3),dtype=intype)
  vtemp = np.zeros((3),dtype=intype)
  vtemp2 = np.zeros((3),dtype=intype)
  r = np.zeros((3),dtype=intype)
  w = numba.float64(0.0)
  for i in numba.prange(n):
    w = q[0]
    r[0] = pf * q[1]
    r[1] = pf * q[2]
    r[2] = pf * q[3]

    # Wow oh wow, numba hates function calls in a loop! 100x speedup
    #vtemp = np.cross(r,v[i,:])
    #vtemp = qv_crossp(r,v[i,:])
    vtemp[0] = r[1] * v[i,2] - r[2] * v[i,1]
    vtemp[1] = r[2] * v[i,0] - r[0] * v[i,2]
    vtemp[2] = r[0] * v[i,1] - r[1] * v[i,0]

    for j in range(3):
      vtemp[j] += w * v[i,j]

    #vtemp2 = np.cross(r,vtemp)
    #vtemp2 = qv_crossp(r,vtemp)
    vtemp2[0] = r[1] * vtemp[2] - r[2] * vtemp[1]
    vtemp2[1] = r[2] * vtemp[0] - r[0] * vtemp[2]
    vtemp2[2] = r[0] * vtemp[1] - r[1] * vtemp[0]


    for j in range(3):
      vout[i,j] = 2.0 * vtemp2[j] + v[i,j]

  return vout

@numba.jit(nopython=True,fastmath=nbFastmath,cache=nbcache,parallel=False)
def quat_vectorLN1(q,v,n,intype,p=P):
  pf = numba.float32(p > 0) * 2.0 - 1.0
  vout = np.zeros((n,3),dtype=intype)
  vtemp = np.zeros((3),dtype=intype)
  vtemp2 = np.zeros((3),dtype=intype)
  r = np.zeros((3),dtype=intype)
  w = numba.float64(0.0)
  for i in numba.prange(n):
    w = q[i,0]
    r[0] = pf * q[i,1]
    r[1] = pf * q[i,2]
    r[2] = pf * q[i,3]

    # Wow oh wow, numba hates function calls in a loop! 100x speedup
    # vtemp = np.cross(r,v)
    # vtemp = qv_crossp(r,v)
    vtemp[0] = r[1] * v[2] - r[2] * v[1]
    vtemp[1] = r[2] * v[0] - r[0] * v[2]
    vtemp[2] = r[0] * v[1] - r[1] * v[0]

    for j in range(3):
      vtemp[j] += w * v[j]

    # vtemp2 = np.cross(r,vtemp)
    # vtemp2 = qv_crossp(r,vtemp)
    vtemp2[0] = r[1] * vtemp[2] - r[2] * vtemp[1]
    vtemp2[1] = r[2] * vtemp[0] - r[0] * vtemp[2]
    vtemp2[2] = r[0] * vtemp[1] - r[1] * vtemp[0]

    for j in range(3):
      vout[i,j] = 2.0 * vtemp2[j] + v[j]

  return vout


@numba.jit(nopython=True,fastmath=nbFastmath,cache=nbcache)
def quat_vector1(q,v,p=P):
  pf = numba.float32(p > 0) * 2.0 - 1.0
  vout = np.zeros_like(v)
  vtemp = np.zeros_like(v)
  vtemp2 = np.zeros_like(v)
  r = np.zeros_like(v)
  w = q[0]
  r[0] = pf * q[1]
  r[1] = pf * q[2]
  r[2] = pf * q[3]
  #cross product r,v
  vtemp[0] = r[1] * v[2] - r[2] * v[1]
  vtemp[1] = r[2] * v[0] - r[0] * v[2]
  vtemp[2] = r[0] * v[1] - r[1] * v[0]
  for j in range(3):
    vtemp[j] += w * v[j]

  # cross product r, vtemp
  vtemp2[0] = r[1] * vtemp[2] - r[2] * vtemp[1]
  vtemp2[1] = r[2] * vtemp[0] - r[0] * vtemp[2]
  vtemp2[2] = r[0] * vtemp[1] - r[1] * vtemp[0]

  for j in range(3):
    vout[j] = 2.0 * vtemp2[j] + v[j]

  return vout

def randomquat(n=1):
  aprime = np.pi**(2.0/3.0)
  cu_ran = (np.random.random((n, 3))-0.5)*aprime
  return cu2qu(cu_ran)
# 
def axnorm(ax):
  '''axis-angle normalization -- unit vector length, corrects negative angles and angles over 2Pi'''
  axout = axnormL(np.require(ax,requirements=['C','A']))
  axout = np.squeeze(axout)
  return axout


@numba.jit(nopython=True,fastmath=nbFastmath,cache=nbcache,parallel=nbParallel)
def axnormL(ax):
  axin,m,n,intype = prepIn(ax)
  axout = np.zeros((n,4),dtype=intype)
  for i in numba.prange(n):
    sign = 1.0 if (axin[i,3] >= eps) else -1.0
    axout[i,3] = sign*axin[i,3]
    sum = 0.0
    for j in range(3):
      axout[i,j] = sign * axin[i,j]
      sum += axout[i,j]*axout[i,j]
    axout[i,3] = np.fmod(axout[i,3],PI2)
    sum = 1.0/(np.sqrt(sum)+1e-35)
    for j in range(3):
      axout[i,j] *= sum
    if axout[i,3] > (PI+1e-6):
      for j in range(3):
        axout[i,j] *= -1.0
      axout[i,3] = PI2 - axout[i,3]
  return axout



# This software was developed by employees of the US Naval Research Laboratory (NRL), an
# agency of the Federal Government. Pursuant to title 17 section 105 of the United States
# Code, works of NRL employees are not subject to copyright protection, and this software
# is in the public domain. PyEBSDIndex is an experimental system. NRL assumes no
# responsibility whatsoever for its use by other parties, and makes no guarantees,
# expressed or implied, about its quality, reliability, or any other characteristic. We
# would appreciate acknowledgment if the software is used. To the extent that NRL may hold
# copyright in countries other than the United States, you are hereby granted the
# non-exclusive irrevocable and unconditional right to print, publish, prepare derivative
# works and distribute this software, in any medium, or authorize others to do so on your
# behalf, on a royalty-free basis throughout the world. You may improve, modify, and
# create derivative works of the software or any portion of the software, and you may copy
# and distribute such modifications or works. Modified works should carry a notice stating
# that you changed the software and should note the date and nature of any such change.
# Please explicitly acknowledge the US Naval Research Laboratory as the original source.
# This software can be redistributed and/or modified freely provided that any derivative
# works bear some notice that they are derived from it, and any modified versions bear
# some notice that they have been modified.
#
# Author: David Rowenhorst;
# The US Naval Research Laboratory Date: 21 Aug 2020

"""Setup and handling of Hough indexing runs of EBSD patterns on a
single thread.
"""

from timeit import default_timer as timer

import numpy as np
import h5py

from pyebsdindex import (
    band_vote,
    ebsd_pattern,
    rotlib,
    tripletlib,
    _pyopencl_installed,
)

if _pyopencl_installed:
    from pyebsdindex.opencl import band_detect_cl as band_detect
else:
    from pyebsdindex import band_detect as band_detect


RADEG = 180.0 / np.pi


def index_pats(
    patsin=None,
    filename=None,
    phaselist=["FCC"],
    vendor=None,
    PC=None,
    sampleTilt=70.0,
    camElev=5.3,
    bandDetectPlan=None,
    nRho=90,
    nTheta=180,
    tSigma=None,
    rSigma=None,
    rhoMaskFrac=0.1,
    nBands=9,
    backgroundSub=False,
    patstart=0,
    npats=-1,
    return_indexer_obj=False,
    ebsd_indexer_obj=None,
    clparams=None,
    verbose=0,
    chunksize=528,
    gpu_id=None,
):
    """Index EBSD patterns on a single thread.

    Parameters
    ----------
    patsin : numpy.ndarray, optional
        EBSD patterns in an array of shape (n points, n pattern
        rows, n pattern columns). If not given, these are read from
        ``filename``.
    filename : str, optional
        Name of file with EBSD patterns. If not given, ``patsin`` must
        be passed.
    phaselist : list of str, optional
        Options are ``"FCC"`` and ``"BCC"``. Default is ``["FCC"]``.
    vendor : str, optional
        Which vendor convention to use for the pattern center (PC) and
        the returned orientations. The available options are ``"EDAX"``
        (default), ``"BRUKER"``, ``"OXFORD"``, ``"EMSOFT"``,
        ``"KIKUCHIPY"``.
    PC : list, optional
        Pattern center (PCx, PCy, PCz) in the :attr:`indexer.vendor` or
        ``vendor`` convention. For EDAX TSL, this is (x*, y*, z*),
        defined in fractions of pattern width with respect to the lower
        left corner of the detector. If not passed, this is set to (x*,
        y*, z*) = (0.471659, 0.675044, 0.630139). If
        ``vendor="EMSOFT"``, the PC must be four numbers, the final
        number being the pixel size.
    sampleTilt : float, optional
        Sample tilt towards the detector in degrees. Default is 70
        degrees. Unused if ``ebsd_indexer_obj`` is passed.
    camElev : float, optional
        Camera elevation in degrees. Default is 5.3 degrees. Unused
        if ``ebsd_indexer_obj`` is passed.
    bandDetectPlan : pyebsdindex.band_detect.BandDetect, optional
        Collection of parameters using in band detection. Unused if
        ``ebsd_indexer_obj`` is passed.
    nRho : int, optional
        Default is 90 degrees. Unused if ``ebsd_indexer_obj`` is
        passed.
    nTheta : int, optional
        Default is 180 degrees. Unused if ``ebsd_indexer_obj`` is
        passed.
    tSigma : float, optional
        Unused if ``ebsd_indexer_obj`` is passed.
    rSigma : float, optional
        Unused if ``ebsd_indexer_obj`` is passed.
    rhoMaskFrac : float, optional
        Default is 0.1. Unused if ``ebsd_indexer_obj`` is passed.
    nBands : int, optional
        Number of detected bands to use in triplet voting. Default
        is 9. Unused if ``ebsd_indexer_obj`` is passed.
    backgroundSub : bool, optional
        Whether to subtract a static background prior to indexing.
        Default is ``False``.
    patstart : int, optional
        Starting index of the patterns to index. Default is ``0``.
    npats : int, optional
        Number of patterns to index. Default is ``-1``, which will
        index up to the final pattern in ``patsin``.
    return_indexer_obj : bool, optional
        Whether to return the EBSD indexer. Default is ``False``.
    ebsd_indexer_obj : EBSDIndexer, optional
        EBSD indexer. If not given, many of the above parameters must be
        passed. Otherwise, these parameters are retrieved from this
        indexer.
    clparams : list, optional
        OpenCL parameters passed to :mod:`pyopencl` if the package is
        installed.
    verbose : int, optional
        0 - no output (default), 1 - timings, 2 - timings and the Hough
        transform of the first pattern with detected bands highlighted.
    chunksize : int, optional
        Default is 528.
    gpu_id : int, optional
        ID of GPU to use if :mod:`pyopencl` is installed.

    Returns
    -------
    indxData : numpy.ndarray
        Complex numpy array (or array of structured data), that is
        [nphases + 1, npoints]. The data is stored for each phase used
        in indexing and the ``indxData[-1]`` layer uses the best guess
        on which is the most likely phase, based on the fit, and number
        of bands matched for each phase. Each data entry contains the
        orientation expressed as a quaternion (quat) (using the
        convention of ``vendor`` or :attr:`indexer.vendor`), Pattern
        Quality (pq), Confidence Metric (cm), Phase ID (phase), Fit
        (fit) and Number of Bands Matched (nmatch). There are some other
        metrics reported, but these are mostly for debugging purposes.
    bandData : numpy.ndarray
        Band identification data from the Radon transform.
    indexer : EBSDIndexer
        EBSD indexer, returned if ``return_indexer_obj=True``.
    """
    pats = None
    if patsin is None:
        pdim = None
    else:
        if isinstance(patsin, ebsd_pattern.EBSDPatterns):
            pats = patsin.patterns
        elif isinstance(patsin, np.ndarray):
            pats = patsin
        elif isinstance(patsin, h5py.Dataset):
            shp = patsin.shape
            if len(shp) == 3:
                pats = patsin
            elif len(shp) == 2:  # Just read off disk now
                pats = patsin[()]
                pats = pats.reshape(1, shp[0], shp[1])
        else:
            raise ValueError("Unrecognized input data type")

        pdim = pats.shape[-2:]

    if ebsd_indexer_obj is None:
        indexer = EBSDIndexer(
            filename=filename,
            phaselist=phaselist,
            vendor=vendor,
            PC=PC,
            sampleTilt=sampleTilt,
            camElev=camElev,
            bandDetectPlan=bandDetectPlan,
            nRho=nRho,
            nTheta=nTheta,
            tSigma=tSigma,
            rSigma=rSigma,
            rhoMaskFrac=rhoMaskFrac,
            nBands=nBands,
            patDim=pdim,
            gpu_id=gpu_id,
        )
    else:
        indexer = ebsd_indexer_obj

    if filename is not None:
        indexer.update_file(filename)
    if pats is not None and not np.all(indexer.bandDetectPlan.patDim == np.array(pdim)):
        indexer.update_file(patDim=pats.shape[-2:])

    if backgroundSub:
        indexer.bandDetectPlan.collect_background(
            fileobj=indexer.fID, patsIn=pats, nsample=1000
        )

    dataout, banddata, indxstart, indxend = indexer.index_pats(
        patsin=pats,
        patstart=patstart,
        npats=npats,
        clparams=clparams,
        verbose=verbose,
        chunksize=chunksize,
    )

    if not return_indexer_obj:
        return dataout, banddata
    else:
        return dataout, banddata, indexer


class EBSDIndexer:
    """Setup of Hough indexing of EBSD patterns.

    Parameters
    ----------
    filename : str, optional
        Name of file with EBSD patterns.
    phaselist : list of str, optional
        Options are ``"FCC"`` and ``"BCC"``. Default is ``["FCC"]``.
    vendor : str, optional
        Which vendor convention to use for the pattern center (PC) and
        the returned orientations. The available options are ``"EDAX"``
        (default), ``"BRUKER"``, ``"OXFORD"``, ``"EMSOFT"``,
        ``"KIKUCHIPY"``.
    PC : list, optional
        Pattern center (PCx, PCy, PCz) in the ``vendor`` convention. For
        EDAX TSL, this is (x*, y*, z*), defined in fractions of pattern
        width with respect to the lower left corner of the detector. If
        not passed, this is set to (x*, y*, z*) = (0.471659, 0.675044,
        0.630139). If ``vendor="EMSOFT"``, the PC must be four numbers,
        the final number being the pixel size.
    sampleTilt : float, optional
        Sample tilt towards the detector in degrees. Default is 70
        degrees. Unused if ``ebsd_indexer_obj`` is passed.
    camElev : float, optional
        Camera elevation in degrees. Default is 5.3 degrees. Unused
        if ``ebsd_indexer_obj`` is passed.
    bandDetectPlan : pyebsdindex.band_detect.BandDetect, optional
        Collection of parameters using in band detection. Unused if
        ``ebsd_indexer_obj`` is passed.
    nRho : int, optional
        Default is 90 degrees. Unused if ``ebsd_indexer_obj`` is
        passed.
    nTheta : int, optional
        Default is 180 degrees. Unused if ``ebsd_indexer_obj`` is
        passed.
    tSigma : float, optional
        Unused if ``ebsd_indexer_obj`` is passed.
    rSigma : float, optional
        Unused if ``ebsd_indexer_obj`` is passed.
    rhoMaskFrac : float, optional
        Default is 0.1. Unused if ``ebsd_indexer_obj`` is passed.
    nBands : int, optional
        Number of detected bands to use in triplet voting. Default
        is 9. Unused if ``ebsd_indexer_obj`` is passed.
    patDim : int, optional
        Number of dimensions of pattern array.
    **kwargs
        Keyword arguments passed on to
        :class:`~pyebsdindex.band_detect.BandDetect`.
    """

    def __init__(
        self,
        filename=None,
        phaselist=["FCC"],
        vendor=None,
        PC=None,
        sampleTilt=70.0,
        camElev=5.3,
        bandDetectPlan=None,
        nRho=90,
        nTheta=180,
        tSigma=1.0,
        rSigma=1.2,
        rhoMaskFrac=0.15,
        nBands=9,
        patDim=None,
        **kwargs
    ):
        """Create an EBSD indexer."""
        self.filein = filename
        if self.filein is not None:
            self.fID = ebsd_pattern.get_pattern_file_obj(self.filein)
        else:
            self.fID = None

        self.phaselist = phaselist
        self.phaseLib = []
        for ph in self.phaselist:
            self.phaseLib.append(band_vote.BandVote(tripletlib.triplib(libType=ph)))

        self.vendor = "EDAX"
        if vendor is None:
            if self.fID is not None:
                self.vendor = self.fID.vendor
        else:
            self.vendor = vendor

        if PC is None:
            self.PC = np.array([0.471659, 0.675044, 0.630139])  # A default value
        else:
            self.PC = np.asarray(PC)

        self.PCcorrectMethod = None
        self.PCcorrectParam = None

        self.sampleTilt = sampleTilt
        self.camElev = camElev

        if bandDetectPlan is None:
            self.bandDetectPlan = band_detect.BandDetect(
                nRho=nRho,
                nTheta=nTheta,
                tSigma=tSigma,
                rSigma=rSigma,
                rhoMaskFrac=rhoMaskFrac,
                nBands=nBands,
                **kwargs
            )
        else:
            self.bandDetectPlan = bandDetectPlan

        if self.fID is not None:
            self.bandDetectPlan.band_detect_setup(
                patDim=[self.fID.patternW, self.fID.patternH]
            )
        elif patDim is not None:
            self.bandDetectPlan.band_detect_setup(patDim=patDim)

        self.dataTemplate = np.dtype(
            [
                ("quat", np.float64, 4),
                ("iq", np.float32),
                ("pq", np.float32),
                ("cm", np.float32),
                ("phase", np.int32),
                ("fit", np.float32),
                ("nmatch", np.int32),
                ("matchattempts", np.int32, 2),
                ("totvotes", np.int32),
            ]
        )

    def update_file(self, filename=None, patDim=np.array([120, 120], dtype=np.int32)):
        """Update file with patterns to index.

        Parameters
        ----------
        filename : str, optional
            Name of file with EBSD patterns.
        patDim : numpy.ndarray
            1D array with two values, the pattern height and width.
        """
        if filename is None:
            self.filein = None
            self.bandDetectPlan.band_detect_setup(patDim=patDim)
        else:
            self.filein = filename
            self.fID = ebsd_pattern.get_pattern_file_obj(self.filein)
            self.bandDetectPlan.band_detect_setup(
                patDim=[self.fID.patternH, self.fID.patternW]
            )

    def index_pats(
        self,
        patsin=None,
        patstart=0,
        npats=-1,
        clparams=None,
        PC=None,
        verbose=0,
        chunksize=528,
    ):
        """Index EBSD patterns.

        Parameters
        ----------
        patsin : numpy.ndarray, optional
            EBSD patterns in an array of shape (n points, n pattern
            rows, n pattern columns). If not given, these are read from
            :attr:`self.filename`.
        patstart : int, optional
            Starting index of the patterns to index. Default is ``0``.
        npats : int, optional
            Number of patterns to index. Default is ``-1``, which will
            index up to the final pattern in ``patsin``.
        clparams : list, optional
            OpenCL parameters passed to :mod:`pyopencl`.
        PC : list, optional
            Pattern center (PC) parameters (PCx, PCy, PCz) in the vendor
            convention. For EDAX TSL, this is (x*, y*, z*), defined in
            fractions of pattern width with respect to the lower left
            corner of the detector. If not given, this is read from
            :attr:`self.PC`. If :attr:`vendor` is ``"EMSOFT"``, the PC
            must be four numbers, the final number being the pixel size.
        verbose : int, optional
            0 - no output (default), 1 - timings, 2 - timings and the
            Hough transform of the first pattern with detected bands
            highlighted.
        chunksize : int, optional
            Default is 528.

        Returns
        -------
        indxData : numpy.ndarray
            Complex numpy array (or array of structured data), that is
            [nphases + 1, npoints]. The data is stored for each phase
            used in indexing and the `indxData[-1]` layer uses the best
            guess on which is the most likely phase, based on the fit,
            and number of bands matched for each phase. Each data entry
            contains the orientation expressed as a quaternion (quat)
            (using `self.vendor`'s convention), Pattern Quality (pq),
            Confidence Metric (cm), Phase ID (phase), Fit (fit) and
            Number of Bands Matched (nmatch). There are some other
            metrics reported, but these are mostly for debugging
            purposes.
        bandData : numpy.ndarray
            Band identification data from the Radon transform.
        patstart : int
            Starting index of the indexed patterns.
        npats : int
            Number of patterns indexed. This and `patstart` are useful
            for the distributed indexing procedures.
        """
        if patsin is None:
            pats = self.fID.read_data(
                returnArrayOnly=True,
                patStartCount=[patstart, npats],
                convertToFloat=True,
            )
        else:
            pshape = patsin.shape
            if len(pshape) == 2:
                pats = np.reshape(patsin, (1, pshape[0], pshape[1]))
            else:
                pats = patsin  # [patStart:patEnd, :,:]
            pshape = pats.shape

            if self.bandDetectPlan.patDim is None:
                self.bandDetectPlan.band_detect_setup(patDim=pshape[1:3])
            else:
                if np.all((np.array(pshape[1:3]) - self.bandDetectPlan.patDim) == 0):
                    self.bandDetectPlan.band_detect_setup(patDim=pshape[1:3])

        if self.bandDetectPlan.patDim is None:
            self.bandDetectPlan.band_detect_setup(patterns=pats)

        npoints = pats.shape[0]
        if npats == -1:
            npats = npoints

        bandData = self.bandDetectPlan.find_bands(
            pats, clparams=clparams, verbose=verbose, chunksize=chunksize
        )
        shpBandDat = bandData.shape
        if PC is None:
            PC_0 = self.PC
        else:
            PC_0 = PC
        bandNorm = self.bandDetectPlan.radonPlan.radon2pole(
            bandData, PC=PC_0, vendor=self.vendor
        )

        # Return bandNorm, patStart, patEnd
        tic = timer()
        nPhases = len(self.phaseLib)
        q = np.zeros((nPhases, npoints, 4))
        indxData = np.zeros((nPhases + 1, npoints), dtype=self.dataTemplate)

        indxData["phase"] = -1
        indxData["fit"] = 180.0
        indxData["totvotes"] = 0
        earlyexit = max(7, shpBandDat[1])
        for i in range(npoints):
            bandNorm1 = bandNorm[i, :, :]
            bDat1 = bandData[i, :]
            whgood = np.nonzero(bDat1["max"] > -1.0e6)[0]
            if whgood.size >= 3:
                bDat1 = bDat1[whgood]
                bandNorm1 = bandNorm1[whgood, :]
                indxData["pq"][0:nPhases, i] = np.sum(bDat1["max"], axis=0)

                for j in range(len(self.phaseLib)):
                    (
                        avequat,
                        fit,
                        cm,
                        bandmatch,
                        nMatch,
                        matchAttempts,
                        totvotes,
                    ) = self.phaseLib[j].tripvote(
                        bandNorm1, band_intensity=bDat1["avemax"], verbose=verbose,
                    )
                    # avequat,fit,cm,bandmatch,nMatch, matchAttempts = self.phaseLib[j].pairVoteOrientation(bandNorm1,goNumba=True)
                    if nMatch >= 3:
                        q[j, i, :] = avequat
                        indxData["fit"][j, i] = fit
                        indxData["cm"][j, i] = cm
                        indxData["phase"][j, i] = j
                        indxData["nmatch"][j, i] = nMatch
                        indxData["matchattempts"][j, i] = matchAttempts
                        indxData["totvotes"][j, i] = totvotes
                    if nMatch >= earlyexit:
                        break

        qref2detect = self._refframe2detector()
        q = q.reshape(nPhases * npoints, 4)
        q = rotlib.quat_multiply(q, qref2detect)
        q = rotlib.quatnorm(q)
        q = q.reshape(nPhases, npoints, 4)
        indxData["quat"][0:nPhases, :, :] = q
        if nPhases > 1:
            for j in range(nPhases - 1):
                indxData[-1, :] = np.where(
                    (indxData[j, :]["cm"] * indxData[j, :]["nmatch"])
                    > (indxData[j + 1, :]["cm"] * indxData[j + 1, :]["nmatch"]),
                    indxData[j, :],
                    indxData[j + 1, :],
                )
        else:
            indxData[-1, :] = indxData[0, :]

        if verbose > 0:
            print("Band Vote Time: ", timer() - tic)

        return indxData, bandData, patstart, npats

    def _refframe2detector(self):
        ven = str.upper(self.vendor)
        if ven in ["EDAX", "EMSOFT", "KIKUCHIPY"]:
            q0 = np.array([np.sqrt(2.0) * 0.5, 0.0, 0.0, -1.0 * np.sqrt(2.0) * 0.5])
            tiltang = -1.0 * (90.0 - self.sampleTilt + self.camElev) / RADEG
            q1 = np.array([np.cos(tiltang * 0.5), np.sin(tiltang * 0.5), 0.0, 0.0])
            quatref2detect = rotlib.quat_multiply(q1, q0)
        elif ven in ["OXFORD", "BRUKER"]:
            tiltang = -1.0 * (90.0 - self.sampleTilt + self.camElev) / RADEG
            q1 = np.array([np.cos(tiltang * 0.5), np.sin(tiltang * 0.5), 0.0, 0.0])
            quatref2detect = q1
        else:
            raise ValueError("`self.vendor` unknown")

        return quatref2detect

#    def pcCorrect(self, xy=[[0.0, 0.0]]):
#        # TODO: At somepoint we will put some methods here for
#        #  correcting the PC depending on the location within the scan.
#        #  Need to correct band_detect.radon2pole to accept a PC for
#        #  each point.
#        pass



from pathlib import Path
import numpy as np
from pyebsdindex import rotlib


def writeang(filename, indexer, data,
             gridtype = 'SqrGrid', xstep=1.0, ystep=1.0,
             ncols = None, nrows=None):
  fpath = Path(filename).expanduser()
  with open(fpath,'w',encoding = 'utf-8') as f:
    f.write('# HEADER: Start \r\n')
    f.write('# TEM_PIXperUM          1.000000\r\n')
    f.write('# x-star                ' + str(indexer.PC[0])+'\r\n')
    f.write('# y-star                ' + str(indexer.PC[1])+'\r\n')
    f.write('# z-star                ' + str(indexer.PC[2])+'\r\n')
    f.write('# SampleTiltAngle       ' + str(indexer.sampleTilt)+'\r\n')
    f.write('# CameraElevationAngle  ' + str(indexer.camElev)+'\r\n')
    f.write('# '+'\r\n')
    pcount = 1
    nphase = len(indexer.phaseLib)
    for phase in reversed(indexer.phaseLib):
      f.write('# Phase '+str(nphase - pcount + 1)+'\r\n')
      f.write('# MaterialName \t' + str(phase.phase_name)+'\r\n')
      f.write('# Formula '+'\t \r\n')
      f.write('# Info '+'\t\t \r\n')
      f.write('# Symmetry              '+str(phase.tripLib.laue_code)+'\r\n')
      f.write('# PointGroupID              ' + str(phase.tripLib.symmetry_pgID)+'\r\n')
      f.write('# LatticeConstants      '+ ' '.join(str(' {:.3f}'.format(x)) for x in phase.tripLib.latticeParameter)+'\r\n')
      f.write('# NumberFamilies             ' + str(phase.tripLib.nfamily)+'\r\n')
      for i in range(phase.tripLib.nfamily):
        f.write('# hklFamilies   \t' + (' '.join(str(x).rjust(2,' ') for x in phase.tripLib.family[i,:])) + ' 1 0.00000 1'+'\r\n')
      f.write('# '+'\r\n')
      pcount += 1

    f.write('# '+'\r\n')
    f.write('# GRID: '+gridtype+'\r\n')
    if indexer.fID is not None:
      if indexer.fID.xStep is not None:
        xstep = indexer.fID.xStep
        ystep = indexer.fID.yStep

    f.write('# XSTEP: ' + str(xstep)+'\r\n')
    f.write('# YSTEP: ' + str(ystep)+'\r\n')
    if ncols is None:
      ncols = 1
      nrows = data.shape[-1]
      if indexer.fID is not None:
        if indexer.fID.nCols is not None:
          ncols = indexer.fID.nCols
          nrows = indexer.fID.nRows
    else:
      if nrows is None:
        nrows = np.ceil(data.shape[-1]/ncols)


    ncols = int(ncols)
    nrows = int(nrows)
    f.write('# NCOLS_ODD: ' + str(ncols)+'\r\n')
    f.write('# NCOLS_EVEN: ' + str(ncols)+'\r\n')
    f.write('# NROWS: ' + str(nrows)+'\r\n')
    f.write('# VERSION 7'+'\r\n')
    f.write('# COLUMN_COUNT: 10'+'\r\n')
    f.write('# HEADER: End'+'\r\n')

    nphase = data.shape[0]-1
    if nphase == 1:
      phaseIDadd = 0
    else:
      phaseIDadd = 1
    eulers = rotlib.qu2eu(data[-1]['quat'])
    for i in range(data.shape[-1]):
      line = '  '
      line += '   '.join('{:.5f}'.format(x) for x in eulers[i,:])
      line += ' '
      line += ('{:.5f}'.format((i % ncols)*float(xstep))).rjust(12,' ') + ' '
      line += ('{:.5f}'.format((int(i / ncols)) * float(ystep))).rjust(12, ' ') + ' '
      line += ('{:.1f}'.format(data[-1]['pq'][i])).rjust(8, ' ') + ' '
      if data[-1]['phase'][i] < 0: # this is an unindexed point.
        phase = 0
        ci = -1.0
        fit = 0.00
      else:
        phase = data[-1]['phase'][i]+phaseIDadd
        ci = data[-1]['cm'][i]
        fit = data[-1]['fit'][i]

      line += '{:.3f}'.format(ci).rjust(6, ' ') + ' '
      line += ' {:}'.format(phase) + ''
      line += '1'.rjust(7, ' ')+''
      line += ('{:.3f}'.format(fit)).rjust(7, ' ')
      f.write(line+'\r\n')


'''This software was developed by employees of the US Naval Research Laboratory (NRL), an
agency of the Federal Government. Pursuant to title 17 section 105 of the United States
Code, works of NRL employees are not subject to copyright protection, and this software
is in the public domain. PyEBSDIndex is an experimental system. NRL assumes no
responsibility whatsoever for its use by other parties, and makes no guarantees,
expressed or implied, about its quality, reliability, or any other characteristic. We
would appreciate acknowledgment if the software is used. To the extent that NRL may hold
copyright in countries other than the United States, you are hereby granted the
non-exclusive irrevocable and unconditional right to print, publish, prepare derivative
works and distribute this software, in any medium, or authorize others to do so on your
behalf, on a royalty-free basis throughout the world. You may improve, modify, and
create derivative works of the software or any portion of the software, and you may copy
and distribute such modifications or works. Modified works should carry a notice stating
that you changed the software and should note the date and nature of any such change.
Please explicitly acknowledge the US Naval Research Laboratory as the original source.
This software can be redistributed and/or modified freely provided that any derivative
works bear some notice that they are derived from it, and any modified versions bear
some notice that they have been modified.

Author: David Rowenhorst;
The US Naval Research Laboratory Date: 21 Aug 2020'''


from pathlib import Path
from timeit import default_timer as timer

import numba
import numpy as np
import scipy.optimize as opt
import matplotlib.pyplot as plt

from pyebsdindex import ebsd_pattern

#from os import environ
#environ["NUMBA_CACHE_DIR"] = str(tempdir)


class NLPAR():
  def __init__(self, filename=None,  lam=0.7, searchradius=3,dthresh=0.0, nrows = None, ncols = None):
    self.lam = lam
    self.searchradius = searchradius
    self.dthresh = dthresh
    self.filepath = None
    self.hdfdatapath = None
    self.filepathout = None
    self.hdfdatapathout = None
    #self.patternfile = None
    #self.patternfileout = None
    self.setfile(filename)
    self.mask = None
    self.sigma = None
    self.nrows = None
    self.ncols = None
    if nrows is not None:
      self.nrows = nrows
    if ncols is not None:
      self.ncols = ncols


  def setfile(self,filepath=None):
    self.filepath = None
    self.hdfdatapath = None
    pathtemp = np.atleast_1d(filepath)
    fpath = pathtemp[0]
    hdf5path = None
    if pathtemp.size > 1:
      hdf5path = pathtemp[1]
    if fpath is not None:
      self.filepath = Path(fpath)
      self.hdfdatapath = hdf5path




  def setoutfile(self,patternfile, filepath=None):
    '''patternfile is an input pattern file object from ebsd_pattern.  Filepath is a string.
    In the future I want to be able to specify the HDF5 data path to store the output data, but that
    is proving to be a bit of a mess.  For now, a copy of the original HDF5 is made, and the NLPAR patterns will be
    overwritten on top of the originals. '''
    self.filepathout = None
    self.hdfdatapathout = None
    pathtemp = np.atleast_1d(filepath)
    fpath = pathtemp[0]
    hdf5path = None
    #if pathtemp.size > 1:
    #  hdf5path = pathtemp[1]
    #print(fpath, hdf5path)
    if fpath is not None: # the user has set an output file path.
      self.filepathout = Path(fpath).expanduser().resolve()
      self.hdfdatapathout =  hdf5path
      if patternfile.filetype != 'HDF5': #check that the input and output are not the same.
        pathok = self.filepathout.exists()
        if pathok:
          pathok = not self.filepathout.samefile(patternfile.filepath)
          if not pathok:
            raise ValueError('Error: File input and output are exactly the same.')
            return

        patternfile.copy_file([self.filepathout,self.hdfdatapathout] )
        return  # fpath and (maybe) hdf5 path were set manually.
      else: # this is a hdf5 file
        if self.hdfdatapathout is None:
          patternfile.copy_file(self.filepathout)
          self.hdfdatapathout = patternfile.h5patdatpth
          return
        else:
          patternfile.copy_file([self.filepathout, self.hdfdatapathout])
          return

    if patternfile is not None: # the user has set no path.
      hdf5path = None
      if patternfile.filetype == 'UP':
        p = Path(patternfile.filepath)
        appnd = "_NLPAR_l{:1.2f}".format(self.lam) + "sr{:d}".format(self.searchradius)
        newfilepath = str(p.parent / Path(p.stem + appnd + p.suffix))
        patternfile.copy_file(newfilepath)

      if patternfile.filetype == 'HDF5':
        hdf5path_tmp = str(patternfile.h5patdatpth).split('/')
        if hdf5path_tmp[0] == '':
          hdf5path_org =  hdf5path_tmp[1]
        else:
          hdf5path_org = hdf5path_tmp[0]
        p = Path(patternfile.filepath)
        appnd = "_NLPAR_l{:1.2f}".format(self.lam) + "sr{:d}".format(self.searchradius)
        hdf5path = hdf5path_org+appnd
        newfilepath = str(p.parent / Path(p.stem + appnd + p.suffix))
        #patternfile.copy_file([newfilepath, hdf5path_org], newh5path=hdf5path)
        patternfile.copy_file([newfilepath])
        hdf5path = patternfile.h5patdatpth

      self.filepathout = newfilepath
      self.hdfdatapathout = hdf5path
      return

  def getinfileobj(self):
    if self.filepath is not None:
      fID = ebsd_pattern.get_pattern_file_obj([self.filepath, self.hdfdatapath])
      if fID.nRows is not None:
        self.nrows = fID.nRows
      else:
        fID.nRows = self.nrows
      if fID.nCols is not None:
        self.ncols = fID.nCols
      else:
        fID.nCols = self.ncols
      return fID
    else:
      return None

  def getoutfileobj(self):
    if self.filepathout is not None:
      return ebsd_pattern.get_pattern_file_obj([self.filepathout, self.hdfdatapathout])
    else:
      return None

  def opt_lambda(self,chunksize=0,saturation_protect=True,automask=True, backsub = False,
                 target_weights=[0.5, 0.34, 0.25], dthresh=0.0, autoupdate=True):

    target_weights = np.asarray(target_weights)

    def loptfunc(lam,d2,tw,dthresh):
      temp = (d2 > dthresh).choose(dthresh, d2)
      dw = np.exp(-(temp) / lam ** 2)
      w = np.sum(dw, axis=2) + 1e-12

      metric = np.mean(np.abs(tw - 1.0 / w))
      return metric


    @numba.njit(fastmath=True, cache=True,parallel=True)
    def d2norm(d2, n2, dij, sigma):
      shp = d2.shape
      s2 = sigma**2
      for j in numba.prange(shp[0]):
        for i in range(shp[1]):
          for q in range(shp[2]):
            if n2[j,i,q] > 0:
              ii = dij[j,i,q,1]
              jj = dij[j,i,q,0]
              s2_12 = (s2[j,i]+s2[jj,ii])
              d2[j,i,q] -= n2[j,i,q] * s2_12
              d2[j,i,q] /= s2_12*np.sqrt(2.0*n2[j,i,q])

    patternfile = self.getinfileobj()
    patternfile.read_header()
    nrows = np.uint64(self.nrows) #np.uint64(patternfile.nRows)
    ncols = np.uint64(self.ncols) #np.uint64(patternfile.nCols)

    pwidth = np.uint64(patternfile.patternW)
    pheight = np.uint64(patternfile.patternH)
    phw = pheight * pwidth

    nn = 1
    if chunksize <= 0:
      chunksize = np.maximum(1, np.round(1e9 / phw / ncols) ) # keeps the chunk at about 4GB
      chunksize = np.minimum(nrows,chunksize)
      print("Chunk size set to nrows:", int(chunksize))
    chunksize = np.int64(chunksize)

    nn = np.uint64(nn)

    if (automask is True) and (self.mask is None):
      self.mask = (self.automask(pheight,pwidth))
    if self.mask is None:
      self.mask = np.ones((pheight,pwidth),dtype=np.uint8)

    indices = np.asarray((self.mask.flatten().nonzero())[0],np.uint64)

    sigma = np.zeros((nrows,ncols),dtype=np.float32)+1e24
    colstartcount = np.asarray([0,ncols],dtype=np.int64)


    dthresh = np.float32(dthresh)
    lamopt_values = []
    for j in range(0,nrows,chunksize):
      print('Block',j)
      #rowstartread = np.int64(max(0,j - nn))
      rowstartread = np.int64(j)
      rowend = min(j + chunksize + nn,nrows)
      rowcountread = np.int64(rowend - rowstartread)
      data = patternfile.read_data(patStartCount=[[0,rowstartread],[ncols,rowcountread]],
                                        convertToFloat=True,returnArrayOnly=True)

      shp = data.shape

      if backsub is True:
        data = self.backsub(data)
        #back = np.mean(data, axis=0)
        #back -= np.mean(back)
        #data -= back
      data = data.reshape(shp[0], phw)

      rowstartcount = np.asarray([0,rowcountread],dtype=np.int64)
      sigchunk, (d2,n2, dij) = self.sigma_numba(data,nn,rowcountread,ncols,rowstartcount,colstartcount,indices,saturation_protect)
      tmp = (sigma[j:j + rowstartcount[1],:] < sigchunk).choose( sigchunk, sigma[j:j + rowstartcount[1],:])
      sigma[j:j + rowstartcount[1],:] = tmp


      d2norm(d2, n2, dij, sigchunk)
      lamopt_values_chnk = []
      for tw in target_weights:
        lam = 1.0
        lambopt1 = opt.minimize(loptfunc,lam,args=(d2,tw,dthresh),method='Nelder-Mead',
                                bounds = [[0.001, 10.0]],options={'fatol': 0.0001})
        lamopt_values_chnk.append(lambopt1['x'])


      lamopt_values.append(lamopt_values_chnk)
    lamopt_values = np.asarray(lamopt_values)
    print("Range of lambda values: ", np.mean(lamopt_values, axis = 0).flatten())
    print("Optimal Choice: ", np.median(np.mean(lamopt_values, axis = 0)))
    if autoupdate == True:
      self.lam = np.median(np.mean(lamopt_values, axis = 0))
    if self.sigma is None:
      self.sigma = sigma

  def calcnlpar(self, chunksize=0, searchradius=None, lam = None, dthresh = None, saturation_protect=True, automask=True,
                filename=None, fileout=None, reset_sigma=True, backsub = False, rescale = False):

    if lam is not None:
      self.lam = lam

    if dthresh is not None:
      self.dthresh = dthresh

    if searchradius is not None:
      self.searchradius = searchradius

    lam = np.float32(self.lam)
    dthresh = np.float32(self.dthresh)
    sr = np.int64(self.searchradius)

    if filename is not None:
      self.setfile(filepath=filename)

    if reset_sigma:
      self.sigma = None

    patternfile = self.getinfileobj()

    #if filepathout is not None:
    self.setoutfile(patternfile, filepath=fileout)

    patternfileout = self.getoutfileobj()


    nrows = np.int64(self.nrows)#np.int64(patternfile.nRows)
    ncols = np.int64(self.ncols)#np.int64(patternfile.nCols)
    if patternfileout.nCols is None:
      patternfileout.nCols = ncols
    if patternfileout.nRows is None:
      patternfileout.nRows = nrows


    pwidth = np.int64(patternfile.patternW)
    pheight = np.int64(patternfile.patternH)
    phw = pheight*pwidth

    if chunksize <= 0:
      chunksize = np.maximum(1, np.round(1e9 / phw / ncols) ) # keeps the chunk at about 8GB
      chunksize = np.minimum(nrows, chunksize)
      print("Chunk size set to nrows:", int(chunksize))
    chunksize = np.int64(chunksize)



    if (automask is True) and (self.mask is None):
      self.mask = (self.automask(pheight,pwidth))
    if self.mask is None:
      self.mask = np.ones((pheight,pwidth), dtype=np.uint8)

    indices = np.asarray( (self.mask.flatten().nonzero())[0], np.uint64)
    calcsigma = False
    if self.sigma is None:
      calcsigma = True
      self.sigma = np.zeros((nrows, ncols), dtype=np.float32)+1e24


    if np.asarray(self.sigma).size == 1:
      tmp = np.asarray(self.sigma)[0]
      self.sigma =  np.zeros((nrows, ncols), dtype=np.float32)+tmp
      calcsigma = False

    shpsigma = np.asarray(self.sigma).shape
    if (shpsigma[0] != nrows) and (shpsigma[1] != ncols):
      self.sigma = np.zeros((nrows,ncols),dtype=np.float32) + 1e24
      calcsigma = True


    sigma = np.asarray(self.sigma)
    scalemethod = 'clip'
    if rescale == True:
      if np.issubdtype(patternfileout.filedatatype, np.integer):
        mxval = np.iinfo(patternfileout.filedatatype).max
        scalemethod = 'fullscale'
      else: # not int, so no rescale.
        rescale = False

    nthreadpos = numba.get_num_threads()
    #numba.set_num_threads(36)
    colstartcount = np.asarray([0,ncols],dtype=np.int64)
    print(lam, sr, dthresh)

    for j in range(0,nrows,chunksize):
      rowstartread = np.int64(max(0, j-sr))
      rowend = min(j + chunksize+sr,nrows)
      rowcountread = np.int64(rowend-rowstartread)
      data = patternfile.read_data(patStartCount = [[0,rowstartread], [ncols,rowcountread]],
                                        convertToFloat=True,returnArrayOnly=True)

      shpdata = data.shape

      if backsub is True:
        data = self.backsub(data)


      data = data.reshape(shpdata[0], phw)

      rowstartcount = np.asarray([0,rowcountread],dtype=np.int64)
      if calcsigma is True:
        sigchunk, tmp = self.sigma_numba(data,1,rowcountread,ncols,rowstartcount,colstartcount,indices,saturation_protect)
        del tmp
        tmp = (sigma[rowstartread:rowend,:] < sigchunk).choose(sigchunk,sigma[rowstartread:rowend,:])
        sigma[rowstartread:rowend,:] = tmp
      else:
        sigchunk = sigma[rowstartread:rowend,:]

      print('Block', j)
      dataout = self.nlpar_nb(data,lam, sr, dthresh, sigchunk,
                              rowcountread,ncols,indices,saturation_protect)

      dataout = dataout.reshape(rowcountread, ncols, phw)
      dataout = dataout[j-rowstartread:, :, : ]
      shpout = dataout.shape
      dataout = dataout.reshape(shpout[0]*shpout[1], pheight, pwidth)
      if rescale == True:
        for i in range(dataout.shape[0]):
          temp = dataout[i,:,:]
          temp -= temp.min()
          temp *= float(mxval)/temp.max()
          dataout[i,:,:] = temp

      patternfileout.write_data(newpatterns=dataout,patStartCount = [[0,j], [ncols, shpout[0]]],
                                     flt2int='clip',scalevalue=1.0 )
      #self.patternfileout.write_data(newpatterns=dataout,patStartCount=[j*ncols,shpout[0]*shpout[1]],
      #                               flt2int='clip',scalevalue=1.0 )
      #return dataout
      #sigma[j:j+rowstartcount[1],:] += \
      #  sigchunk[rowstartcount[0]:rowstartcount[0]+rowstartcount[1],:]
    numba.set_num_threads(nthreadpos)


  def calcsigma(self,chunksize=0,nn=1,saturation_protect=True,automask=True):

    patternfile = self.getinfileobj()


    nrows = np.int64(self.nrows)#np.uint64(patternfile.nRows)
    ncols = np.int64(self.ncols)#np.uint64(patternfile.nCols)

    pwidth = np.uint64(patternfile.patternW)
    pheight = np.uint64(patternfile.patternH)
    phw = pheight*pwidth

    if chunksize <= 0:
      chunksize = np.round(2e9/phw/ncols) # keeps the chunk at about 8GB
      chunksize = np.minimum(nrows,chunksize)
      print("Chunk size set to nrows:", int(chunksize))
    chunksize = np.int64(chunksize)


    nn = np.uint64(nn)

    if (automask is True) and (self.mask is None):
      self.mask = (self.automask(pheight,pwidth))
    if self.mask is None:
      self.mask = np.ones((pheight,pwidth), dytype=np.uint8)

    indices = np.asarray( (self.mask.flatten().nonzero())[0], np.uint64)

    sigma = np.zeros((nrows, ncols), dtype=np.float32)
    #d_nn = np.zeros((nrows, ncols, int((2*nn+1)**2)), dtype=np.float32)
    colstartcount = np.asarray([0,ncols],dtype=np.int64)
    dave = 0.0
    for j in range(0,nrows,chunksize):
      rowstartread = np.int64(max(0, j-nn))
      rowend = min(j + chunksize+nn,nrows)
      rowcountread = np.int64(rowend-rowstartread)
      data = patternfile.read_data(patStartCount = [[0,rowstartread], [ncols,rowcountread]],
                                        convertToFloat=True,returnArrayOnly=True)

      shp = data.shape
      data = data.reshape(data.shape[0], phw)

      #data = None
      if rowend == nrows:
        rowstartcount = np.asarray([j-rowstartread,rowcountread - (j-rowstartread) ], dtype=np.int64)
      else:
        rowstartcount = np.asarray([j-rowstartread,chunksize ], dtype=np.int64)
      dtic = timer()
      sigchunk, temp = self.sigma_numba(data,nn, rowcountread,ncols,rowstartcount,colstartcount,indices,saturation_protect)
      dave += (timer() - dtic)
      sigma[j:j+rowstartcount[1],:] += \
        sigchunk[rowstartcount[0]:rowstartcount[0]+rowstartcount[1],:]

    return sigma

  def backsub(self, data):
    # This function will fit a 2D gaussian on top of a plane to the averaged set of patterns (data) that is provided.
    # It will automatically use whatever mask is defined for valid data.
    # If the gaussian fit fails to converge, it will fall back to just using the mean set of patterns for the background
    # with a warning.


    def gaussian_surf(x, y, a, x0, y0, sigx, sigy, c, d, e):
    # equation for 2D gaussian on top of a plane.
      return a * np.exp(- ((x - x0) ** 2) / (2.0 * sigx ** 2) - ((y - y0) ** 2) / (2.0 * sigy ** 2)) + c + d * x + e * y

    def fit_gauss(M, *args):
    # helper function
      x, y = M
      #arr = np.zeros(x.shape)
      return gaussian_surf(x, y, *args)

    back = np.mean(data, axis=0) # start with the mean of all the data
    # now fit a 2D gaussian sitting on a plane.  See fuction def above.
    nx = data.shape[-1]
    ny = data.shape[-2]
    x = np.arange(nx, dtype=float)
    x = (np.broadcast_to(x.reshape(1,nx), (ny, nx)))
    y = np.arange(ny, dtype=float)
    y = (np.broadcast_to(y, (nx, ny)).T)
    x = x.ravel()
    y = y.ravel()

    # need to grab only the values that are in the mask.
    wh = np.nonzero(self.mask.ravel())[0]
    xwh = x[wh]
    ywh = y[wh]
    xywh = np.vstack((xwh, ywh))
    zwh = (back.ravel())[wh]
    whmx = np.unravel_index(back.argmax(), back.shape)
    minz = zwh.min()
    # initialize a guess for the parameters.
    # [gauss amplitude, max loc x, max loc y, sigx, sigy, const offset, slope x, slope y]
    p0 = [(zwh.max() - zwh.min()), whmx[1], whmx[0], nx/2.355, ny/2.355, minz, 0, 0]
    try:
      popt, pcov = opt.curve_fit(fit_gauss, xywh, zwh, p0)
      backfit = (gaussian_surf(x, y, *popt)).reshape(ny, nx)
      #print(p0, popt)
    except RuntimeError:
      print('Warning: no convergence on back subtract ... using mean of the patterns.')
      print('This may not be ideal for scans with few grains across the width of the scan.')
      backfit = back
    backfit -= np.mean(backfit)
    #f, axarr = plt.subplots(1, 3)
    #f.set_size_inches(10, 4)
    #axarr[0].imshow(data[0,:,:].squeeze(), cmap='gray')
    #axarr[1].imshow(data[0,:,:].squeeze() - backfit, cmap='gray')
    #axarr[2].imshow(backfit, cmap='gray')

    data -= backfit
    return data

  @staticmethod
  def automask( h,w ):
    r = (min(h,w)*0.98*0.5)
    x = np.arange(w, dtype=np.float32)
    x = np.minimum(x , (w-x))
    x = x.reshape(1,w)
    y = np.arange(h, dtype=np.float32)
    y = np.minimum(y , (h - y))
    y = y.reshape(h,1)

    mask =np.sqrt(y**2 + x**2)
    mask = (mask < r).astype(np.uint8)
    mask = np.roll(mask, (int(h/2), int(w/2) ), axis=(0,1))

    return mask

  @staticmethod
  @numba.jit(nopython=True,cache=True,fastmath=False,parallel=True)
  def sigma_numba(data, nn, nrows, ncols, rowstartcount, colstartcount, indices, saturation_protect=True):
    sigma = np.zeros((nrows,ncols), dtype = np.float32)
    dout = np.zeros((nrows,ncols,(nn*2+1)**2), dtype=np.float32)
    nout = np.zeros((nrows,ncols,(nn * 2 + 1) ** 2),dtype=np.float32)
    dij = np.zeros((nrows,ncols,(nn * 2 + 1) ** 2, 2),dtype=np.uint64)
    shpdata = data.shape

    n0 = np.float32(shpdata[-1])
    shpind = indices.shape
    mxval = np.max(data)
    if saturation_protect == False:
      mxval += 1.0
    else:
      mxval *= 0.9961

    for j in numba.prange(rowstartcount[0], rowstartcount[0]+rowstartcount[1]):
      nn_r_start = j - nn if (j - nn) >= 0 else 0
      nn_r_end = (j + nn if (j + nn) < nrows else nrows-1)+1
      for i in numba.prange(colstartcount[0], colstartcount[0]+colstartcount[1]):
        nn_c_start = i - nn if (i - nn) >= 0 else 0
        nn_c_end = (i + nn if (i + nn) < ncols else ncols - 1) + 1

        mind = np.float32(1e24)
        indx_0 = i+ncols*j
        count = 0
        for j_nn in range(nn_r_start,nn_r_end ):
          for i_nn in range(nn_c_start,nn_c_end):
            dij[j,i,count,0] = np.uint64(j_nn)
            dij[j,i,count,1] = np.uint64(i_nn) # want to save this for labmda optimization
            indx_nn = i_nn+ncols*j_nn
            d2 = np.float32(0.0)
            n2 = np.float32(0.0)
            nout[j,i,count] = n0 # want to save this for labmda optimization
            if not((i == i_nn) and (j == j_nn)):
              for q in range(shpind[0]):
                d0 = data[indx_0, indices[q]]
                d1 = data[indx_nn, indices[q]]
                if (d1 < mxval) and (d0 < mxval):
                  n2 += 1.0
                  d2 += (d0 - d1)**2
              nout[j,i,count] = n2
              s0 = d2 / np.float32(n2 * 2.0)
              if d2 >= 1.e-3: #sometimes EDAX collects the same pattern twice
                if s0 < mind:
                  mind = s0
            dout[j,i,count] = d2 # want to save this for labmda optimization

            count += 1

        sigma[j,i] = np.sqrt(mind)
    return sigma,( dout, nout, dij)

  @staticmethod
  @numba.jit(nopython=True,cache=True,fastmath=True,parallel=True)
  def nlpar_nb(data,lam, sr, dthresh, sigma, nrows,ncols,indices,saturation_protect=True):

    def getpairid(idx0, idx1):
      idx0_t = int(idx0)
      idx1_t = int(idx1)
      if idx0 < idx1:
        pairid = idx0_t + (idx1_t << 32)
      else:
        pairid = idx1_t + (idx0_t << 32)
      return numba.uint64(pairid)

    lam2 = 1.0 / lam**2
    dataout = np.zeros_like(data, np.float32)
    shpdata = data.shape
    shpind = indices.shape
    winsz = np.int32((2*sr+1)**2)


    mxval = np.max(data)
    if saturation_protect == False:
      mxval += np.float32(1.0)
    else:
      mxval *= np.float32(0.999)
    for i in numba.prange(ncols):
      winstart_x = max((i - sr),0) - max((i + sr - (ncols - 1)),0)
      winend_x = min((i + sr),(ncols - 1)) + max((sr - i),0) + 1
      pairdict = numba.typed.Dict.empty(key_type=numba.core.types.uint64,value_type=numba.core.types.float32)
      for j in range(nrows):
        winstart_y = max((j - sr),0) - max((j + sr - (nrows - 1)),0)
        winend_y = min((j + sr),(nrows - 1)) + max((sr - j),0) + 1

        weights = np.zeros(winsz,dtype=np.float32)
        pindx = np.zeros(winsz,dtype=np.uint64)

        indx_0 = i + ncols * j
        sigma0 = sigma[j,i]**2

        counter = 0
        for j_nn in range(winstart_y,winend_y):
          for i_nn in range(winstart_x,winend_x):
            indx_nn = i_nn + ncols * j_nn
            pindx[counter] = indx_nn

            if indx_nn == indx_0:
              weights[counter] = np.float32(-1.0e6)
            else:
              pairid = getpairid(indx_0, indx_nn)
              if pairid in pairdict:
                weights[counter] = pairdict[pairid]
              else:
                sigma1 = sigma[j_nn,i_nn]**2
                d2 = np.float32(0.0)
                n2 = np.float32(0.0)
                for q in range(shpind[0]):
                  d0 = data[indx_0,indices[q]]
                  d1 = data[indx_nn,indices[q]]
                  if (d1 < mxval) and (d0 < mxval):
                    n2 += np.float32(1.0)
                    d2 += (d0 - d1) ** np.int32(2)
                d2 -= n2*(sigma0+sigma1)
                dnorm = (sigma1 + sigma0)*np.sqrt(np.float32(2.0)*n2)
                if dnorm > np.float32(1.e-8):
                  d2 /= dnorm
                else:
                  d2 = np.float32(1e6)*n2
                weights[counter] = d2
                pairdict[pairid] = numba.float32(d2)
            counter += 1
        #print('________________')
        # end of window scanning
        sum = np.float(0.0)
        for i_nn in range(winsz):

          weights[i_nn] = np.maximum(weights[i_nn]-dthresh, numba.float32(0.0))
          weights[i_nn] = np.exp(-1.0 * weights[i_nn] * lam2)
          sum += weights[i_nn]

        for i_nn in range(winsz):
          indx_nn = pindx[i_nn]
          weights[i_nn] /= sum
          #print(weights[i_nn], ' \n')
          for q in range(shpdata[1]):
            dataout[indx_0, q] += data[indx_nn, q]*weights[i_nn]
        #print('_______', '\n')
    return dataout





# Generation of a dynamically simulated Al 20 kV EBSD pattern.
import kikuchipy as kp
import numpy as np
from orix.quaternion import Rotation
import skimage.io as skio


mp = kp.load(
    "/home/hakon/kode/emsoft/emdata/crystal_data/al/al_mc_mp_20kv.h5",
    projection="lambert",
    energy=20
)
rot = Rotation.identity()
detector = kp.detectors.EBSDDetector(
    shape=(100, 120),
    pc=(0.4, 0.6, 0.5),
    convention="tsl",
    sample_tilt=70,
    tilt=5.3,
)
s = mp.get_patterns(
    rotations=rot,
    detector=detector,
    compute=True,
    energy=20,
    dtype_out=np.uint8,
)
skio.imsave(
    "/home/hakon/kode/pyebsdindex/pyebsdindex/tests/data/al_sim_20kv/al_sim_20kv.png",
    s.data.squeeze()
)



'''This software was developed by employees of the US Naval Research Laboratory (NRL), an
agency of the Federal Government. Pursuant to title 17 section 105 of the United States
Code, works of NRL employees are not subject to copyright protection, and this software
is in the public domain. PyEBSDIndex is an experimental system. NRL assumes no
responsibility whatsoever for its use by other parties, and makes no guarantees,
expressed or implied, about its quality, reliability, or any other characteristic. We
would appreciate acknowledgment if the software is used. To the extent that NRL may hold
copyright in countries other than the United States, you are hereby granted the
non-exclusive irrevocable and unconditional right to print, publish, prepare derivative
works and distribute this software, in any medium, or authorize others to do so on your
behalf, on a royalty-free basis throughout the world. You may improve, modify, and
create derivative works of the software or any portion of the software, and you may copy
and distribute such modifications or works. Modified works should carry a notice stating
that you changed the software and should note the date and nature of any such change.
Please explicitly acknowledge the US Naval Research Laboratory as the original source.
This software can be redistributed and/or modified freely provided that any derivative
works bear some notice that they are derived from it, and any modified versions bear
some notice that they have been modified.

Author: David Rowenhorst;
The US Naval Research Laboratory Date: 21 Aug 2020'''


import numpy as np
from os import path
import pyopencl as cl
from os import environ
environ['PYOPENCL_COMPILER_OUTPUT'] = '1'

RADDEG = 180.0/np.pi
DEGRAD = np.pi/180.0



class OpenClParam():
  def __init__(self, gpu_id=0):
    self.platform = None
    self.gpu = None
    self.ngpu = 0
    self.gpu_id = gpu_id
    self.ctx = None
    self.prg = None
    self.queue = None
    self.memflags = cl.mem_flags

    try:
      self.get_context()

    except Exception as e:
      if hasattr(e,'message'):
        print(e.message)
      else:
        print(e)

  def get_platform(self):
    self.platform = cl.get_platforms()[0]
  def get_gpu(self):

    if self.platform is None:
      self.get_platform()

    self.gpu = self.platform.get_devices(device_type=cl.device_type.GPU)
    self.ngpu = len(self.gpu)
    if len(self.gpu)-1 < self.gpu_id:
      self.gpu_id = len(self.gpu)-1

  def get_context(self):
    if self.gpu is None:
      self.get_gpu()

    self.ctx = cl.Context(devices = self.gpu)
    kernel_location = path.dirname(__file__)
    self.prg = cl.Program(self.ctx,open(path.join(kernel_location,'clkernels.cl')).read()).build()

  def get_queue(self, gpu_id=None, random_gpu=False):

    if self.ctx is None:
      self.get_context()

    if gpu_id is None:
      gpu_id = self.gpu_id

    if random_gpu == True:
      gpu_id = np.random.randint(len(self.gpu))
    gpuindx = min(len(self.gpu)-1, gpu_id)
    self.gpu_id = gpuindx
    self.queue = cl.CommandQueue(self.ctx, device=self.gpu[gpuindx])





from timeit import default_timer as timer

import numpy as np

from pyebsdindex.rotlib import *
from pyebsdindex.rotlib import quatnorm, quatconj, quat_multiply


pi = np.pi
RADDEG = 180.0/pi
# I had some issues with the periodic boundaries in cubochoric/homochoric/quaternion spaces, so I am using
# quaternion misorientation to evaluate the error.

def testrotlib(float32=False, return_quat=False, seed = 1, n = 1000000):
  n = np.int(n)
  np.random.seed(seed)
  qu = (np.random.random((n,4))*2.0-1)
  if float32 is True:
    qu = qu.astype(np.float32)
  qu = quatnorm(qu)
  qu[0, :] = np.array([1.0, 0.0, 0.0, 0.0])
  qu[1, :] = np.array([0.0,0.0,0.0,-1.0])
  qu[2, :] = np.array([0.0,  0.4264014489670335 , -0.6396021436482270, -0.6396021436482270])

  totaltime = 0.0
  print( "qu2: ")
  tic = timer()
  qu1 = ax2qu(qu2ax(qu))
  t1 = timer() - tic
  totaltime += t1
  mx = np.arccos(quatnorm(quat_multiply(qu,quatconj(qu1)))[:,0]) * 2*RADDEG
  print("    ax: ", str(mx.max()), str(mx.mean()) )
  print("        Time to execute: ", t1)  
  tic = timer()
  qu1 = cu2qu(qu2cu(qu))
  t1 = timer() - tic
  totaltime += t1
  mx = np.arccos(quatnorm(quat_multiply(qu,quatconj(qu1)))[:,0]) * 2*RADDEG
  print("    cu: ", str(mx.max()), str(mx.mean()) )
  print("        Time to execute: ", t1)  
  tic = timer()
  qu1 = ho2qu(qu2ho(qu))
  t1 = timer() - tic
  totaltime += t1
  mx = np.arccos(quatnorm(quat_multiply(qu,quatconj(qu1)))[:,0]) * 2*RADDEG
  print("    ho: ", str(mx.max()), str(mx.mean()) )
  print("        Time to execute: ", t1)  
  tic = timer()
  qu1 = om2qu(qu2om(qu))
  t1 = timer() - tic
  totaltime += t1
  mx = np.arccos(quatnorm(quat_multiply(qu,quatconj(qu1)))[:,0]) * 2*RADDEG
  print("    om: ", str(mx.max()), str(mx.mean()) )
  print("        Time to execute: ", t1)  
  tic = timer()
  qu1 = eu2qu(qu2eu(qu))
  t1 = timer() - tic
  totaltime += t1
  mx = np.arccos(quatnorm(quat_multiply(qu,quatconj(qu1)))[:,0]) * 2*RADDEG
  print("    eu: ", str(mx.max()), str(mx.mean()) )
  print("        Time to execute: ", t1)  
  tic = timer()
  qu1 = ro2qu(qu2ro(qu))
  t1 = timer() - tic
  totaltime += t1
  mx = np.arccos(quatnorm(quat_multiply(qu,quatconj(qu1)))[:,0]) * 2*RADDEG
  print("    ro: ", str(mx.max()), str(mx.mean()) )
  print("        Time to execute: ", t1)

  eu = qu2eu(qu)
  print("eu2: ")
  tic = timer()
  qu1 = eu2qu(ax2eu(eu2ax(eu)))
  t1 = timer() - tic
  totaltime += t1
  mx = np.arccos(quatnorm(quat_multiply(qu,quatconj(qu1)))[:,0]) * 2*RADDEG
  print("    ax: ",str(mx.max()), str(mx.mean()))
  print("        Time to execute: ", t1)  
  tic = timer()
  qu1 = eu2qu(cu2eu(eu2cu(eu)))
  t1 = timer() - tic
  totaltime += t1
  mx = np.arccos(quatnorm(quat_multiply(qu,quatconj(qu1)))[:,0]) * 2*RADDEG
  print("    cu: ",str(mx.max()), str(mx.mean()))
  print("        Time to execute: ", t1)  
  tic = timer()
  qu1 = eu2qu(ho2eu(eu2ho(eu)))
  t1 = timer() - tic
  totaltime += t1
  mx = np.arccos(quatnorm(quat_multiply(qu,quatconj(qu1)))[:,0]) * 2*RADDEG
  print("    ho: ",str(mx.max()), str(mx.mean()))
  print("        Time to execute: ", t1)  
  tic = timer()
  qu1 = eu2qu(om2eu(eu2om(eu)))
  t1 = timer() - tic
  totaltime += t1
  mx = np.arccos(quatnorm(quat_multiply(qu,quatconj(qu1)))[:,0]) * 2*RADDEG
  print("    om: ",str(mx.max()), str(mx.mean()))
  print("        Time to execute: ", t1)  
  tic = timer()
  qu1 = eu2qu(ro2eu(eu2ro(eu)))
  t1 = timer() - tic
  totaltime += t1
  mx = np.arccos(quatnorm(quat_multiply(qu,quatconj(qu1)))[:,0]) * 2*RADDEG
  print("    ro: ",str(mx.max()), str(mx.mean()))
  print("        Time to execute: ", t1)

  ax = qu2ax(qu)
  print("ax2: ")
  tic = timer()
  qu1 = ax2qu(cu2ax(ax2cu(ax)))
  t1 = timer() - tic
  totaltime += t1
  mx = np.arccos(quatnorm(quat_multiply(qu,quatconj(qu1)))[:,0]) * 2*RADDEG
  print("    cu: ",str(mx.max()), str(mx.mean()))
  print("        Time to execute: ", t1)  
  tic = timer()
  qu1 = ax2qu(ho2ax(ax2ho(ax)))
  t1 = timer() - tic
  totaltime += t1
  mx = np.arccos(quatnorm(quat_multiply(qu,quatconj(qu1)))[:,0]) * 2*RADDEG
  print("    ho: ",str(mx.max()), str(mx.mean()))
  print("        Time to execute: ", t1)  
  tic = timer()
  qu1 = ax2qu(om2ax(ax2om(ax)))
  t1 = timer() - tic
  totaltime += t1
  mx = np.arccos(quatnorm(quat_multiply(qu,quatconj(qu1)))[:,0]) * 2*RADDEG
  print("    om: ",str(mx.max()), str(mx.mean()))
  print("        Time to execute: ", t1)  
  tic = timer()
  qu1 = ax2qu(ro2ax(ax2ro(ax)))
  t1 = timer() - tic
  totaltime += t1
  mx = np.arccos(quatnorm(quat_multiply(qu,quatconj(qu1)))[:,0]) * 2*RADDEG
  print("    ro: ",str(mx.max()), str(mx.mean()))
  print("        Time to execute: ", t1)

  cu = qu2cu(qu)
  print("cu2: ")
  tic = timer()
  qu1 = cu2qu(ho2cu(cu2ho(cu)))
  t1 = timer() - tic
  totaltime += t1
  mx = np.arccos(quatnorm(quat_multiply(qu,quatconj(qu1)))[:,0]) * 2*RADDEG
  print("    ho: ",str(mx.max()), str(mx.mean()))
  print("        Time to execute: ", t1)  
  tic = timer()
  qu1 = cu2qu(om2cu(cu2om(cu)))
  t1 = timer() - tic
  totaltime += t1
  mx = np.arccos(quatnorm(quat_multiply(qu,quatconj(qu1)))[:,0]) * 2*RADDEG
  print("    om: ",str(mx.max()), str(mx.mean()))
  print("        Time to execute: ", t1)  
  tic = timer()
  qu1 = cu2qu(ro2cu(cu2ro(cu)))
  t1 = timer() - tic
  totaltime += t1
  mx = np.arccos(quatnorm(quat_multiply(qu,quatconj(qu1)))[:,0]) * 2*RADDEG
  print("    ro: ",str(mx.max()), str(mx.mean()))
  print("        Time to execute: ", t1)

  ho = qu2ho(qu)
  print("ho2: ")
  tic = timer()
  qu1 = ho2qu(om2ho(ho2om(ho)))
  t1 = timer() - tic
  totaltime += t1
  mx = np.arccos(quatnorm(quat_multiply(qu,quatconj(qu1)))[:,0]) * 2*RADDEG
  print("    om: ",str(mx.max()), str(mx.mean()))
  print("        Time to execute: ", t1)  
  tic = timer()
  qu1 = ho2qu(ro2ho(ho2ro(ho)))
  t1 = timer() - tic
  totaltime += t1
  mx = np.arccos(quatnorm(quat_multiply(qu,quatconj(qu1)))[:,0]) * 2*RADDEG
  print("    ro: ",str(mx.max()), str(mx.mean()))
  print("        Time to execute: ", t1)

  om = qu2om(qu)
  print("om2: ")
  tic = timer()
  qu1 = om2qu(ro2om(om2ro(om)))
  t1 = timer() - tic
  totaltime += t1
  mx = np.arccos(quatnorm(quat_multiply(qu,quatconj(qu1)))[:,0]) * 2*RADDEG
  print("    ro: ",str(mx.max()), str(mx.mean()))
  print("        Time to execute: ", t1)

  print("Total execution time:", totaltime)

  if return_quat is True:
    return qu



__author__ = "Dave Rowenhorst"
__author_email__ = ""
# Initial committer first, then sorted by line contributions
__credits__ = [
    "Dave Rowenhorst",
    "Håkon Wiik Ånes",
]
__description__ = "Python based tool for Hough/Radon based EBSD indexing"
__name__ = "pyebsdindex"
__version__ = "0.1.1"


# Try to import only once
try:
    import pyopencl
    _pyopencl_installed = True
except ImportError:
    _pyopencl_installed = False

try:
    import ray
    _ray_installed = True
except ImportError:
    _ray_installed = False
